# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

"""
Computation of parcellations using a hierarchical approach.
Author: Bertrand Thirion, 2008
"""

import numpy as np
from numpy.random import rand

from nipy.algorithms.clustering.utils import kmeans, voronoi
from .parcellation import MultiSubjectParcellation
from nipy.algorithms.graph.field import Field
from nipy.algorithms.graph.graph import wgraph_from_coo_matrix


def _jointly_reduce_data(data1, data2, chunksize):
    lnvox = data1.shape[0]
    aux = np.argsort(rand(lnvox)) [:np.minimum(chunksize, lnvox)]
    rdata1 = data1[aux]
    rdata2 = data2[aux]
    return rdata1, rdata2


def _reduce_and_concatenate(data1, data2, chunksize):
    nb_subj = len(data1)
    rset1 = []
    rset2 = []
    for s in range(nb_subj):
        rdata1, rdata2 = _jointly_reduce_data(data1[s], data2[s],
                                              chunksize / nb_subj)
        rset1.append(rdata1)
        rset2.append(rdata2)
    rset1 = np.concatenate(rset1)
    rset2 = np.concatenate(rset2)
    return rset1, rset2


def _field_gradient_jac(ref, target):
    """
    Given a reference field ref and a target field target
    compute the jacobian of the target with respect to ref

    Parameters
    ----------
    ref: Field instance
        that yields the topology of the space
    target : array of shape(ref.V,dim)

    Returns
    -------
    fgj: array of shape (ref.V)
        that gives the jacobian implied by the ref.field->target transformation.
    """
    import numpy.linalg as nl
    n = ref.V
    xyz = ref.field
    dim = xyz.shape[1]
    fgj = []
    ln = ref.list_of_neighbors()
    for i in range(n):
        j = ln[i]
        if np.size(j) > dim - 1:
            dx = np.squeeze(xyz[j] - xyz[i])
            df = np.squeeze(target[j] - target[i])
            FG = np.dot(nl.pinv(dx), df)
            fgj.append(nl.det(FG))
        else:
            fgj.append(1)

    fgj = np.array(fgj)
    return fgj


def _exclusion_map_dep(i, ref, target, targeti):
    """ ancillary function to determine admissible values of some position
    within some predefined values

    Parameters
    ----------
    i (int): index of the structure under consideration
    ref: Field that represent the topological structure of parcels
         and their standard position
    target: array of shape (ref.V,3): current posistion of the parcels
    targeti array of shape (n,3): possible new positions for the ith item

    Returns
    -------
    emap: aray of shape (n): a potential that yields the fitness
          of the proposed positions given the current configuration
    rmin (double): ancillary parameter
    """
    xyz = ref.field
    ln = ref.list_of_neighbors()
    j = ln[i]
    if np.size(j) > 0:
        dx = xyz[j] - xyz[i]
        dx = np.squeeze(dx)
        rmin = np.min(np.sum(dx ** 2, 1)) / 4
        u0 = xyz[i] + np.mean(target[j] - xyz[j], 1)
        emap = - np.sum((targeti - u0) ** 2, 1) + rmin
    else:
        emap = np.zeros(targeti.shape[0])
    return emap


def _exclusion_map(i, ref, target, targeti):
    """Ancillary function to determin admissible values of some position
    within some predefined values

    Parameters
    ----------
    i (int): index of the structure under consideration
    ref: Field that represent the topological structure of parcels
         and their standard position
    target= array of shape (ref.V,3): current posistion of the parcels
    targeti array of shape (n,3): possible new positions for the ith item

    Returns
    -------
    emap: aray of shape (n): a potential that yields the fitness
          of the proposed positions given the current configuration
    rmin (double): ancillary parameter
    """
    xyz = ref.field
    fd = target.shape[1]
    ln = ref.list_of_neighbors()
    j = ln[i]
    j = np.reshape(j, np.size(j))
    rmin = 0
    if np.size(j) > 0:
        dx = np.reshape(xyz[j] - xyz[i], (np.size(j), fd))
        rmin = np.mean(np.sum(dx ** 2, 1)) / 4
        u0 = xyz[i] + np.mean(target[j] - xyz[j], 0)
        emap = rmin - np.sum((targeti - u0) ** 2, 1)
        for k in j:
            amap = np.sum((targeti - target[k]) ** 2, 1) - rmin / 4
            emap[amap < 0] = amap[amap < 0]
    else:
        emap = np.zeros(targeti.shape[0])
    return emap, rmin


def _field_gradient_jac_Map_(i, ref, target, targeti):
    """
    Given a reference field ref and a target field target
    compute the jacobian of the target with respect to ref
    """
    import scipy.linalg as nl
    xyz = ref.field
    fgj = []
    ln = ref.list_of_neighbors()
    j = ln[i]
    if np.size(j) > 0:
        dx = xyz[j] - xyz[i]
        dx = np.squeeze(dx)
        idx = nl.pinv(dx)
        for k in range(targeti.shape[0]):
            df = target[j] - targeti[k]
            df = np.squeeze(df)
            fg = np.dot(idx, df)
            fgj.append(nl.det(fg))
    else:
        fgj = np.zeros(targeti.shape[0])

    fgj = np.array(fgj)
    return fgj


def _field_gradient_jac_Map(i, ref, target, targeti):
    """
    Given a reference field ref and a target field target
    compute the jacobian of the target with respect to ref
    """
    import scipy.linalg as nl
    xyz = ref.field
    fgj = []
    ln = ref.list_of_neighbors()
    j = ln[i]
    if np.size(j) > 0:
        dx = xyz[j] - xyz[i]
        dx = np.squeeze(dx)
        idx = nl.pinv(dx)
        for k in range(targeti.shape[0]):
            df = target[j] - targeti[k]
            df = np.squeeze(df)
            fg = np.dot(idx, df)
            fgj.append(nl.det(fg))
        fgj = np.array(fgj)

        for ij in np.squeeze(j):
            aux = []
            jj = np.squeeze(ln[ij])
            dx = xyz[jj] - xyz[ij]
            dx = np.squeeze(dx)
            idx = nl.pinv(dx)
            ji = np.nonzero(jj == i)
            for k in range(targeti.shape[0]):
                df = target[jj] - target[ij]
                df[ji] = targeti[k] - target[ij]
                df = np.squeeze(df)
                fg = np.dot(idx, df)
                aux.append(nl.det(fg))
            aux = np.array(aux)
            fgj = np.minimum(fgj, aux)
    else:
        fgj = np.zeros(targeti.shape[0])
    return fgj


def _optim_hparcel(feature, domain, graphs, nb_parcel, lamb=1., dmax=10.,
                   niter=5, initial_mask=None, chunksize=1.e5, verbose=0):
    """ Core function of the heirrachical parcellation procedure.

    Parameters
    ----------
    feature: list of subject-related feature arrays
    Pa : parcellation instance that is updated
    graphs: graph that represents the topology of the parcellation
    anat_coord: array of shape (nvox,3) space defining set of coordinates
    nb_parcel: int
               the number of desrired parcels
    lamb=1.0: parameter to weight position
              and feature impact on the algorithm
    dmax = 10: locality parameter (in the space of anat_coord)
              to limit surch volume (CPU save)
    chunksize = int, optional
    niter = 5: number of iterations in the algorithm
    verbose=0: verbosity level

    Returns
    -------
    U: list of arrays of length nsubj
       subject-dependent parcellations
    Proto_anat: array of shape (nvox) labelling of the common space
                (template parcellation)
    """
    nb_subj = len(feature)

    # a1. perform a rough clustering of the data to make prototype
    indiv_coord = np.array([domain.coord[initial_mask[:, s] > - 1]
                             for s in range(nb_subj)])
    reduced_anat, reduced_feature = _reduce_and_concatenate(
        indiv_coord, feature, chunksize)

    _, labs, _ = kmeans(reduced_feature, nb_parcel, Labels=None, maxiter=10)
    proto_anat = [np.mean(reduced_anat[labs == k], 0)
                  for k in range(nb_parcel)]
    proto_anat = np.array(proto_anat)
    proto = [np.mean(reduced_feature[labs == k], 0) for k in range(nb_parcel)]
    proto = np.array(proto)

    # a2. topological model of the parcellation
    # group-level part
    spatial_proto = Field(nb_parcel)
    spatial_proto.set_field(proto_anat)
    spatial_proto.voronoi_diagram(proto_anat, domain.coord)
    spatial_proto.set_gaussian(proto_anat)
    spatial_proto.normalize()

    for git in range(niter):
        LP = []
        LPA = []
        U = []
        Energy = 0
        for s in range(nb_subj):
            # b.subject-specific instances of the model
            # b.0 subject-specific information
            Fs = feature[s]
            lac = indiv_coord[s]
            target = proto_anat.copy()
            lseeds = np.zeros(nb_parcel, np.int)
            aux = np.argsort(rand(nb_parcel))
            toto = np.zeros(lac.shape[0])
            for j in range(nb_parcel):
                # b.1 speed-up :only take a small ball
                i = aux[j]
                dx = lac - target[i]
                iz = np.nonzero(np.sum(dx ** 2, 1) < dmax ** 2)
                iz = np.reshape(iz, np.size(iz))
                if np.size(iz) == 0:
                    iz = np.array([np.argmin(np.sum(dx ** 2, 1))])

                # b.2: anatomical constraints
                lanat = np.reshape(lac[iz], (np.size(iz),
                                             domain.coord.shape[1]))
                pot = np.zeros(np.size(iz))
                JM, rmin = _exclusion_map(i, spatial_proto, target, lanat)
                pot[JM < 0] = np.inf
                pot[JM >= 0] = - JM[JM >= 0]

                # b.3: add feature discrepancy
                df = Fs[iz] - proto[i]
                df = np.reshape(df, (np.size(iz), proto.shape[1]))
                pot += lamb * np.sum(df ** 2, 1)

                # b.4: solution
                if np.sum(np.isinf(pot)) == np.size(pot):
                    pot = np.sum(dx[iz] ** 2, 1)

                sol = iz[np.argmin(pot)]
                target[i] = lac[sol]
                lseeds[i] = sol
                toto[sol] = 1

            if verbose > 1:
                jm = _field_gradient_jac(spatial_proto, target)
                print jm.min(), jm.max(), np.sum(toto > 0)

            # c.subject-specific parcellation
            g = graphs[s]
            f = Field(g.V, g.edges, g.weights, Fs)
            U.append(f.constrained_voronoi(lseeds))

            Energy += np.sum((Fs - proto[U[-1]]) ** 2) / \
                np.sum(initial_mask[:, s] > - 1)
            # recompute the prototypes
            # (average in subject s)
            lproto = [np.mean(Fs[U[-1] == k], 0) for k in range(nb_parcel)]
            lproto = np.array(lproto)
            lproto_anat = np.array([np.mean(lac[U[-1] == k], 0)
                                    for k in range(nb_parcel)])

            LP.append(lproto)
            LPA.append(lproto_anat)

        # recompute the prototypes across subjects
        proto_mem = proto.copy()
        proto = np.mean(np.array(LP), 0)
        proto_anat = np.mean(np.array(LPA), 0)
        displ = np.sqrt(np.sum((proto_mem - proto) ** 2, 1).max())
        if verbose:
            print 'energy', Energy, 'displacement', displ

        # recompute the topological model
        spatial_proto.set_field(proto_anat)
        spatial_proto.voronoi_diagram(proto_anat, domain.coord)
        spatial_proto.set_gaussian(proto_anat)
        spatial_proto.normalize()

        if displ < 1.e-4 * dmax:
            break
    return U, proto_anat


def hparcel(domain, ldata, nb_parcel, nb_perm=0, niter=5, mu=10., dmax=10.,
            lamb=100.0, chunksize=1.e5, verbose=0, initial_mask=None):
    """
    Function that performs the parcellation by optimizing the
    inter-subject similarity while retaining the connectedness
    within subject and some consistency across subjects.

    Parameters
    ----------
    domain: discrete_domain.DiscreteDomain instance,
            yields all the spatial information on the parcelled domain
    ldata: list of (n_subj) arrays of shape (domain.size, dim)
           the feature data used to inform the parcellation
    nb_parcel: int,
               the number of parcels
    nb_perm: int, optional,
             the number of times the parcellation and prfx
             computation is performed on sign-swaped data
    niter: int, optional,
           number of iterations to obtain the convergence of the method
           information in the clustering algorithm
    mu: float, optional,
        relative weight of anatomical information
    dmax: float optional,
          radius of allowed deformations
    lamb: float optional
          parameter to control the relative importance of space vs function
    chunksize; int, optional
               number of points used in internal sub-sampling
    verbose: bool, optional,
             verbosity mode
    initial_mask: array of shape (domain.size, nb_subj), optional
                  initial subject-depedent masking of the domain

    Returns
    -------
    Pa: the resulting parcellation structure appended with the labelling
    """
    # a various parameters
    nbvox = domain.size
    nb_subj = len(ldata)
    if initial_mask is None:
        initial_mask = np.ones((nbvox, nb_subj), np.int)

    graphs = []
    feature = []

    for s in range(nb_subj):
        # build subject-specific models of the data
        lnvox = np.sum(initial_mask[:, s] > - 1)
        lac = domain.coord[initial_mask[:, s] > - 1]
        beta = np.reshape(ldata[s], (lnvox, ldata[s].shape[1]))
        lf = np.hstack((beta, mu * lac / (1.e-15 + np.std(domain.coord, 0))))
        feature.append(lf)
        g = wgraph_from_coo_matrix(domain.topology)
        g.remove_trivial_edges()
        graphs.append(g)

    # main function
    all_labels, proto_anat = _optim_hparcel(
        feature, domain, graphs, nb_parcel, lamb, dmax, niter, initial_mask,
        chunksize=chunksize, verbose=verbose)

    # write the individual labelling
    labels = - np.ones((nbvox, nb_subj)).astype(np.int)
    for s in range(nb_subj):
        labels[initial_mask[:, s] > -1, s] = all_labels[s]

    # compute the group-level labels
    template_labels = voronoi(domain.coord, proto_anat)

    # create the parcellation
    pcl = MultiSubjectParcellation(domain, individual_labels=labels,
                                   template_labels=template_labels,
                                   nb_parcel=nb_parcel)
    pcl.make_feature('functional', np.rollaxis(np.array(ldata), 1, 0))

    if nb_perm > 0:
        prfx0 = perm_prfx(domain, graphs, feature, nb_parcel, ldata,
                          initial_mask, nb_perm, niter, dmax, lamb, chunksize)
        return pcl, prfx0
    else:
        return pcl


def perm_prfx(domain, graphs, features, nb_parcel, ldata, initial_mask=None,
              nb_perm=100, niter=5, dmax=10., lamb=100.0, chunksize=1.e5,
              verbose=1):
    """
    caveat: assumes that the functional dimension is 1
    """
    from ..utils.reproducibility_measures import ttest
    # permutations for the assesment of the results
    prfx0 = []
    adim = domain.coord.shape[1]
    nb_subj = len(ldata)
    for q in range(nb_perm):
        feature = []
        sldata = []
        for s in range(nb_subj):
            lf = features[s].copy()
            swap = (rand() > 0.5) * 2 - 1
            lf[:, 0:-adim] = swap * lf[:, 0:-adim]
            sldata.append(swap * ldata[s])
            feature.append(lf)

        # optimization part
        all_labels, proto_anat = _optim_hparcel(
            feature, domain, graphs, nb_parcel, lamb, dmax, niter,
            initial_mask, chunksize=chunksize)
        labels = - np.ones((domain.size, nb_subj)).astype(np.int)
        for s in range(nb_subj):
            labels[initial_mask[:, s] > -1, s] = all_labels[s]

        # compute the group-level labels
        template_labels = voronoi(domain.coord, proto_anat)

        # create the parcellation
        pcl = MultiSubjectParcellation(domain, individual_labels=labels,
                                       template_labels=template_labels)
        pdata = pcl.make_feature('functional',
                                 np.rollaxis(np.array(ldata), 1, 0))
        prfx = ttest(np.squeeze(pdata))
        if verbose:
            print q, prfx.max(0)
        prfx0.append(prfx.max(0))

    return prfx0



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Test the glm utilities.
"""
from __future__ import with_statement

import numpy as np

from nibabel import load, Nifti1Image, save

from ..glm import GeneralLinearModel, data_scaling, FMRILinearModel

from nose.tools import assert_true, assert_equal, assert_raises
from numpy.testing import (assert_array_almost_equal, assert_almost_equal,
                           assert_array_equal)
from nibabel.tmpdirs import InTemporaryDirectory

from nipy.testing import funcfile


def write_fake_fmri_data(shapes, rk=3, affine=np.eye(4)):
    mask_file, fmri_files, design_files = 'mask.nii', [], []
    for i, shape in enumerate(shapes):
        fmri_files.append('fmri_run%d.nii' %i)
        data = 100 + np.random.randn(*shape)
        data[0] -= 10
        save(Nifti1Image(data, affine), fmri_files[-1])
        design_files.append('dmtx_%d.npz' %i)
        np.savez(design_files[-1], np.random.randn(shape[3], rk))
    save(Nifti1Image((np.random.rand(*shape[:3]) > .5).astype(np.int8), 
                     affine), mask_file)
    return mask_file, fmri_files, design_files 


def generate_fake_fmri_data(shapes, rk=3, affine=np.eye(4)):
    fmri_data = []
    design_matrices = []
    for i, shape in enumerate(shapes):
        data = 100 + np.random.randn(*shape)
        data[0] -= 10
        fmri_data.append(Nifti1Image(data, affine))
        design_matrices.append(np.random.randn(shape[3], rk))
    mask = Nifti1Image((np.random.rand(*shape[:3]) > .5).astype(np.int8),
                       affine)
    return mask, fmri_data, design_matrices


def test_high_level_glm_with_paths():
    shapes, rk = ((5, 6, 4, 20), (5, 6, 4, 19)), 3
    with InTemporaryDirectory():
        mask_file, fmri_files, design_files = write_fake_fmri_data(shapes, rk)
        multi_session_model = FMRILinearModel(fmri_files, design_files,
                                              mask_file)
        multi_session_model.fit()
        z_image, = multi_session_model.contrast([np.eye(rk)[1]] * 2)
        assert_array_equal(z_image.get_affine(), load(mask_file).get_affine())
        assert_true(z_image.get_data().std() < 3.)
        # Delete objects attached to files to avoid WindowsError when deleting
        # temporary directory
        del z_image, fmri_files, multi_session_model


def test_high_level_glm_with_data():
    shapes, rk = ((7, 6, 5, 20), (7, 6, 5, 19)), 3
    mask, fmri_data, design_matrices = write_fake_fmri_data(shapes, rk)

    # without mask
    multi_session_model = FMRILinearModel(fmri_data, design_matrices, mask=None)
    multi_session_model.fit()
    z_image, = multi_session_model.contrast([np.eye(rk)[1]] * 2)
    assert_equal(np.sum(z_image.get_data() == 0), 0)

    # compute the mask
    multi_session_model = FMRILinearModel(fmri_data, design_matrices,
                                          m=0, M=.01, threshold=0.)
    multi_session_model.fit()
    z_image, = multi_session_model.contrast([np.eye(rk)[1]] * 2)
    assert_true(z_image.get_data().std() < 3. )

    # with mask
    multi_session_model = FMRILinearModel(fmri_data, design_matrices, mask)
    multi_session_model.fit()
    z_image, effect_image, variance_image= multi_session_model.contrast(
        [np.eye(rk)[:2]] * 2, output_effects=True, output_variance=True)
    assert_array_equal(z_image.get_data() == 0., load(mask).get_data() == 0.)
    assert_true(
        (variance_image.get_data()[load(mask).get_data() > 0, 0] > .001).all())

    # without scaling
    multi_session_model.fit(do_scaling=False)
    z_image, = multi_session_model.contrast([np.eye(rk)[1]] * 2)
    assert_true(z_image.get_data().std() < 3. )


def test_high_level_glm_contrasts():
    shapes, rk = ((5, 6, 7, 20), (5, 6, 7, 19)), 3
    mask, fmri_data, design_matrices = write_fake_fmri_data(shapes, rk)
    multi_session_model = FMRILinearModel(fmri_data, design_matrices, mask=None)
    multi_session_model.fit()
    z_image, = multi_session_model.contrast([np.eye(rk)[:2]] * 2,
                                            contrast_type='tmin-conjunction')
    z1, = multi_session_model.contrast([np.eye(rk)[:1]] * 2)
    z2, = multi_session_model.contrast([np.eye(rk)[1:2]] * 2)
    assert_true((z_image.get_data() < np.maximum(
        z1.get_data(), z2.get_data())).all())


def test_high_level_glm_null_contrasts():
    shapes, rk = ((5, 6, 7, 20), (5, 6, 7, 19)), 3
    mask, fmri_data, design_matrices = generate_fake_fmri_data(shapes, rk)

    multi_session_model = FMRILinearModel(
        fmri_data, design_matrices, mask=None)
    multi_session_model.fit()
    single_session_model = FMRILinearModel(
        fmri_data[:1], design_matrices[:1], mask=None)
    single_session_model.fit()
    z1, = multi_session_model.contrast([np.eye(rk)[:1], np.zeros((1, rk))])
    z2, = single_session_model.contrast([np.eye(rk)[:1]])
    np.testing.assert_almost_equal(z1.get_data(), z2.get_data())


def ols_glm(n=100, p=80, q=10):
    X, Y = np.random.randn(p, q), np.random.randn(p, n)
    glm = GeneralLinearModel(X)
    glm.fit(Y, 'ols')
    return glm, n, p, q


def ar1_glm(n=100, p=80, q=10):
    X, Y = np.random.randn(p, q), np.random.randn(p, n)
    glm = GeneralLinearModel(X)
    glm.fit(Y, 'ar1')
    return glm, n, p, q


def test_glm_ols():
    mulm, n, p, q = ols_glm()
    assert_array_equal(mulm.labels_, np.zeros(n))
    assert_equal(mulm.results_.keys(), [0.0])
    assert_equal(mulm.results_[0.0].theta.shape, (q, n))
    assert_almost_equal(mulm.results_[0.0].theta.mean(), 0, 1)
    assert_almost_equal(mulm.results_[0.0].theta.var(), 1. / p, 1)


def test_glm_beta():
    mulm, n, p, q = ols_glm()
    assert_equal(mulm.get_beta().shape, (q, n)) 
    assert_equal(mulm.get_beta([0, -1]).shape, (2, n))
    assert_equal(mulm.get_beta(6).shape, (1, n))


def test_glm_mse():
    mulm, n, p, q = ols_glm()
    mse = mulm.get_mse()
    assert_array_almost_equal(mse, np.ones(n), 0) 


def test_glm_logL():
    mulm, n, p, q = ols_glm()
    logL = mulm.get_logL()
    assert_array_almost_equal(logL / n, - p * 1.41 * np.ones(n) / n, 0) 


def test_glm_ar():
    mulm, n, p, q = ar1_glm()
    assert_equal(len(mulm.labels_), n)
    assert_true(len(mulm.results_.keys()) > 1)
    tmp = sum([mulm.results_[key].theta.shape[1]
               for key in mulm.results_.keys()])
    assert_equal(tmp, n)


def test_Tcontrast():
    mulm, n, p, q = ar1_glm()
    cval = np.hstack((1, np.ones(9)))
    z_vals = mulm.contrast(cval).z_score()
    assert_almost_equal(z_vals.mean(), 0, 0)
    assert_almost_equal(z_vals.std(), 1, 0)


def test_Fcontrast_1d():
    mulm, n, p, q = ar1_glm()
    cval = np.hstack((1, np.ones(9)))
    con = mulm.contrast(cval, contrast_type='F')
    z_vals = con.z_score()
    assert_almost_equal(z_vals.mean(), 0, 0)
    assert_almost_equal(z_vals.std(), 1, 0)


def test_Fcontrast_nd():
    mulm, n, p, q = ar1_glm()
    cval = np.eye(q)[:3]
    con = mulm.contrast(cval)
    assert_equal(con.contrast_type, 'F')
    z_vals = con.z_score()
    assert_almost_equal(z_vals.mean(), 0, 0)
    assert_almost_equal(z_vals.std(), 1, 0)


def test_Fcontrast_1d_old():
    mulm, n, p, q = ols_glm()
    cval = np.hstack((1, np.ones(9)))
    con = mulm.contrast(cval, contrast_type='F')
    z_vals = con.z_score()
    assert_almost_equal(z_vals.mean(), 0, 0)
    assert_almost_equal(z_vals.std(), 1, 0)


def test_Fcontrast_nd_ols():
    mulm, n, p, q = ols_glm()
    cval = np.eye(q)[:3]
    con = mulm.contrast(cval)
    assert_equal(con.contrast_type, 'F')
    z_vals = con.z_score()
    assert_almost_equal(z_vals.mean(), 0, 0)
    assert_almost_equal(z_vals.std(), 1, 0)


def test_t_contrast_add():
    mulm, n, p, q = ols_glm()
    c1, c2 = np.eye(q)[0], np.eye(q)[1]
    con = mulm.contrast(c1) + mulm.contrast(c2)
    z_vals = con.z_score()
    assert_almost_equal(z_vals.mean(), 0, 0)
    assert_almost_equal(z_vals.std(), 1, 0)


def test_F_contrast_add():
    mulm, n, p, q = ar1_glm()
    # first test with independent contrast
    c1, c2 = np.eye(q)[:2], np.eye(q)[2:4]
    con = mulm.contrast(c1) + mulm.contrast(c2)
    z_vals = con.z_score()
    assert_almost_equal(z_vals.mean(), 0, 0)
    assert_almost_equal(z_vals.std(), 1, 0)
    # first test with dependent contrast
    con1 = mulm.contrast(c1)
    con2 = mulm.contrast(c1) + mulm.contrast(c1)
    assert_almost_equal(con1.effect * 2, con2.effect)
    assert_almost_equal(con1.variance * 2, con2.variance)
    assert_almost_equal(con1.stat() * 2, con2.stat())


def test_t_contrast_mul():
    mulm, n, p, q = ar1_glm()
    con1 = mulm.contrast(np.eye(q)[0])
    con2 = con1 * 2
    assert_almost_equal(con1.z_score(), con2.z_score())
    assert_almost_equal(con1.effect * 2, con2.effect)


def test_F_contrast_mul():
    mulm, n, p, q = ar1_glm()
    con1 = mulm.contrast(np.eye(q)[:4])
    con2 = con1 * 2
    assert_almost_equal(con1.z_score(), con2.z_score())
    assert_almost_equal(con1.effect * 2, con2.effect)


def test_t_contrast_values():
    mulm, n, p, q = ar1_glm(n=1)
    cval = np.eye(q)[0]
    con = mulm.contrast(cval)
    t_ref = mulm.results_.values()[0].Tcontrast(cval).t
    assert_almost_equal(np.ravel(con.stat()), t_ref)


def test_F_contrast_calues():
    mulm, n, p, q = ar1_glm(n=1)
    cval = np.eye(q)[:3]
    con = mulm.contrast(cval)
    F_ref = mulm.results_.values()[0].Fcontrast(cval).F
    # Note that the values are not strictly equal,
    # this seems to be related to a bug in Mahalanobis
    assert_almost_equal(np.ravel(con.stat()), F_ref, 3)


def test_tmin():
    mulm, n, p, q = ar1_glm(n=1)
    c1, c2, c3 = np.eye(q)[0], np.eye(q)[1], np.eye(q)[2]
    t1, t2, t3 = mulm.contrast(c1).stat(), mulm.contrast(c2).stat(), \
        mulm.contrast(c3).stat()
    tmin = min(t1, t2, t3)
    con = mulm.contrast(np.eye(q)[:3], 'tmin-conjunction')
    assert_equal(con.stat(), tmin)


def test_scaling():
    """Test the scaling function"""
    shape = (400, 10)
    u = np.random.randn(*shape)
    mean = 100 * np.random.rand(shape[1])
    Y = u + mean
    Y, mean_ = data_scaling(Y)
    assert_almost_equal(Y.mean(0), 0)
    assert_almost_equal(mean_, mean, 0)
    assert_true(Y.std() > 1)


def test_fmri_inputs():
    # Test processing of FMRI inputs
    func_img = load(funcfile)
    T = func_img.shape[-1]
    des = np.ones((T, 1))
    des_fname = 'design.npz'
    with InTemporaryDirectory():
        np.savez(des_fname, des)
        for fi in func_img, funcfile:
            for d in des, des_fname:
                fmodel = FMRILinearModel(fi, d, mask='compute')
                fmodel = FMRILinearModel([fi], d, mask=None)
                fmodel = FMRILinearModel(fi, [d], mask=None)
                fmodel = FMRILinearModel([fi], [d], mask=None)
                fmodel = FMRILinearModel([fi, fi], [d, d], mask=None)
                fmodel = FMRILinearModel((fi, fi), (d, d), mask=None)
                assert_raises(ValueError, FMRILinearModel, [fi, fi], d,
                              mask=None)
                assert_raises(ValueError, FMRILinearModel, fi, [d, d],
                              mask=None)


if __name__ == "__main__":
    import nose
    nose.run(argv=['', __file__])



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import os
from distutils import log

# Global variables
LIBS = os.path.realpath('libcstat')


def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration, get_numpy_include_dirs
    from numpy.distutils.system_info import get_info

    config = Configuration('labs', parent_package, top_path)

    # cstat library
    config.add_include_dirs(os.path.join(LIBS,'fff'))
    config.add_include_dirs(os.path.join(LIBS,'wrapper'))
    config.add_include_dirs(get_numpy_include_dirs())

    sources = [os.path.join(LIBS,'fff','*.c')]
    sources.append(os.path.join(LIBS,'wrapper','*.c'))

    # Link with lapack if found on the system

    # XXX: We need to better sort out the use of get_info() for Lapack, because
    # using 'lapack' and 'lapack_opt' returns different things even comparing
    # Ubuntu 8.10 machines on 32 vs 64 bit setups.  On OSX
    # get_info('lapack_opt') does not return the keys: 'libraries' and
    # 'library_dirs', but get_info('lapack') does.
    #
    # For now this code should do the right thing on OSX and linux, but we
    # should ask on the numpy list for clarification on the proper approach.

    # XXX: If you modify these lines, remember to pass the information
    # along to the different .so in the neurospin build system.
    # First, try 'lapack_info', as that seems to provide more details on Linux
    # (both 32 and 64 bits):
    lapack_info = get_info('lapack_opt', 0)
    if 'libraries' not in lapack_info:
        # But on OSX that may not give us what we need, so try with 'lapack'
        # instead.  NOTE: scipy.linalg uses lapack_opt, not 'lapack'...
        lapack_info = get_info('lapack', 0)

    # If no lapack install is found, we use the rescue lapack lite
    # distribution included in the package (sources have been
    # translated to C using f2c)
    if not lapack_info:
        log.warn('No lapack installation found, using lapack lite distribution')
        sources.append(os.path.join(LIBS,'lapack_lite','*.c'))
        library_dirs = []
        libraries = []

    # Best-case scenario: lapack found 
    else:
        library_dirs = lapack_info['library_dirs']
        libraries = lapack_info['libraries']
        if 'include_dirs' in lapack_info:
            config.add_include_dirs(lapack_info['include_dirs'])

    # Information message
    print('LAPACK build options:')
    print('library_dirs: %s ' % library_dirs)
    print('libraries: %s ' % libraries)
    print('lapack_info: %s ' % lapack_info)

    config.add_library('cstat',
                       sources=sources,
                       library_dirs=library_dirs,
                       libraries=libraries,
                       extra_info=lapack_info)

    # Subpackages
    config.add_subpackage('bindings')
    config.add_subpackage('glm')
    config.add_subpackage('group')
    config.add_subpackage('spatial_models')
    config.add_subpackage('utils')
    config.add_subpackage('viz_tools')
    config.add_subpackage('datasets')
    config.add_subpackage('tests')

    config.make_config_py() # installs __config__.py

    return config

if __name__ == '__main__':
    print('This is the wrong setup.py file to run')



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
These are several functions for computing reproducibility measures.
A use script should be appended soon on the repository.

In general thuis proceeds as follows:
The dataset is subject to jacknife subampling ('splitting'),
each subsample being analysed independently.
A reproducibility measure is then derived;

All is used to produce the work described in
Analysis of a large fMRI cohort:
Statistical and methodological issues for group analyses.
Thirion B, Pinel P, Meriaux S, Roche A, Dehaene S, Poline JB.
Neuroimage. 2007 Mar;35(1):105-20.

Bertrand Thirion, 2009-2010
"""

import numpy as np
from nipy.labs.spatial_models.discrete_domain import \
    grid_domain_from_binary_array

# ---------------------------------------------------------
# ----- cluster handling functions ------------------------
# ---------------------------------------------------------


def histo_repro(h):
    """ Given the histogram h, compute a standardized reproducibility measure

    Parameters
    ----------
    h array of shape(xmax+1), the histogram values

    Returns
    -------
    hr, float: the measure
    """
    k = np.size(h) - 1
    if k == 1:
        return 0.
    nf = np.dot(h, np.arange(k + 1)) / k
    if nf == 0:
        return 0.
    n1k = np.arange(1, k + 1)
    res = 1.0 * np.dot(h[1:], n1k * (n1k - 1)) / (k * (k - 1))
    return res / nf


def cluster_threshold(stat_map, domain, th, csize):
    """Perform a thresholding of a map at the cluster-level

    Parameters
    ----------
    stat_map: array of shape(nbvox)
         the input data
    domain: Nifti1Image instance,
          referential- and domain-defining image
    th (float): cluster-forming threshold
    cisze (int>0): cluster size threshold

    Returns
    -------
    binary array of shape (nvox): the binarized thresholded map

    Notes
    -----
    Should be replaced by a more standard function in the future
    """
    if stat_map.shape[0] != domain.size:
        raise ValueError('incompatible dimensions')

    # first build a domain of supra_threshold regions
    thresholded_domain = domain.mask(stat_map > th)

    # get the connected components
    label = thresholded_domain.connected_components()

    binary = - np.ones(domain.size)
    binary[stat_map > th] = label
    nbcc = len(np.unique(label))
    for i in range(nbcc):
        if np.sum(label == i) < csize:
            binary[binary == i] = - 1

    binary = (binary > -1)
    return binary


def get_cluster_position_from_thresholded_map(stat_map, domain, thr=3.0,
                                              csize=10):
    """
    the clusters above thr of size greater than csize in
    18-connectivity are computed

    Parameters
    ----------
    stat_map : array of shape (nbvox),
           map to threshold
    mask: Nifti1Image instance,
          referential- and domain-defining image
    thr: float, optional,
         cluster-forming threshold
    cisze=10: int
              cluster size threshold

    Returns
    -------
    positions array of shape(k,anat_dim):
              the cluster positions in physical coordinates
              where k= number of clusters
              if no such cluster exists, None is returned
    """
    # if no supra-threshold voxel, return
    if (stat_map <= thr).all():
        return None

    # first build a domain of supra_threshold regions
    thresholded_domain = domain.mask(stat_map > thr)

    # get the connected components
    label = thresholded_domain.connected_components()

    # get the coordinates
    coord = thresholded_domain.get_coord()

    # get the barycenters
    baryc = []
    for i in range(label.max() + 1):
        if np.sum(label == i) >= csize:
            baryc.append(np.mean(coord[label == i], 0))

    if len(baryc) == 0:
        return None

    baryc = np.vstack(baryc)
    return baryc


def get_peak_position_from_thresholded_map(stat_map, domain, threshold):
    """The peaks above thr in 18-connectivity are computed

    Parameters
    ----------
    stat_map: array of shape (nbvox): map to threshold
    deomain: referential- and domain-defining image
    thr, float: cluster-forming threshold

    Returns
    -------
    positions array of shape(k,anat_dim):
              the cluster positions in physical coordinates
              where k= number of clusters
              if no such cluster exists, None is returned
    """
    from ..statistical_mapping import get_3d_peaks

    # create an image to represent stat_map
    simage = domain.to_image(data=stat_map)
    
    # extract the peaks
    peaks = get_3d_peaks(simage, threshold=threshold, order_th=2)
    if peaks == None:
        return None

    pos = np.array([p['pos'] for p in peaks])
    return pos


# ---------------------------------------------------------
# ----- data splitting functions ------------------------
# ---------------------------------------------------------


def bootstrap_group(nsubj, ngroups):
    """Split the proposed group into redundant subgroups by bootstrap

    Parameters
    ----------
    nsubj (int) the number of subjects in the population
    ngroups(int) Number of subbgroups to be drawn

    Returns
    -------
    samples: a list of ngroups arrays containing
             the indexes of the subjects in each subgroup
    """
    groupsize = nsubj
    samples = [(groupsize * np.random.rand(groupsize)).astype(np.int)
               for i in range(ngroups)]
    return samples


def split_group(nsubj, ngroups):
    """Split the proposed group into random disjoint subgroups

    Parameters
    ----------
    nsubj (int) the number of subjects to be split
    ngroups(int) Number of subbgroups to be drawn

    Returns
    -------
    samples: a list of ngroups arrays containing
             the indexes of the subjects in each subgroup
    """
    groupsize = int(np.floor(nsubj / ngroups))
    rperm = np.argsort(np.random.rand(nsubj))
    samples = [rperm[i * groupsize: (i + 1) * groupsize]
              for i in range(ngroups)]
    return samples


# ---------------------------------------------------------
# ----- statistic computation -----------------------------
# ---------------------------------------------------------


def conjunction(x, vx, k):
    """Returns a conjunction statistic as the sum of the k lowest t-values

    Parameters
    ----------
    x: array of shape(nrows, ncols),
       effect matrix
    vx: array of shape(nrows, ncols),
        variance matrix
    k: int,
       number of subjects in the conjunction

    Returns
    -------
    t array of shape(nrows): conjunction statistic
    """
    t = np.sort(x / np.sqrt(np.maximum(vx, 1.e-15)))
    cjt = np.sum(t[:, :k], 1)
    return cjt


def ttest(x):
    """Returns the t-test for each row of the data x
    """
    from ..group.onesample import stat
    t = stat(x.T, id='student', axis=0)
    return np.squeeze(t)


def fttest(x, vx):
    """Assuming that x and vx represent a effect and variance estimates,
    returns a cumulated ('fixed effects') t-test of the data over each row

    Parameters
    ----------
    x: array of shape(nrows, ncols): effect matrix
    vx: array of shape(nrows, ncols): variance matrix

    Returns
    -------
    t array of shape(nrows): fixed effect statistics array
    """
    if np.shape(x) != np.shape(vx):
        raise ValueError("incompatible dimensions for x and vx")
    n = x.shape[1]
    t = x / np.sqrt(np.maximum(vx, 1.e-15))
    t = t.mean(1) * np.sqrt(n)
    return t


def mfx_ttest(x, vx):
    """Idem fttest, but returns a mixed-effects statistic

    Parameters
    ----------
    x: array of shape(nrows, ncols): effect matrix
    vx: array of shape(nrows, ncols): variance matrix

    Returns
    -------
    t array of shape(nrows): mixed effect statistics array
    """
    from ..group.onesample import stat_mfx
    t = stat_mfx(x.T, vx.T, id='student_mfx', axis=0)
    return np.squeeze(t)


def voxel_thresholded_ttest(x, threshold):
    """Returns a binary map of the ttest>threshold
    """
    t = ttest(x)
    return t > threshold


def statistics_from_position(target, data, sigma=1.0):
    """ Return a number characterizing how close data is from
    target using a kernel-based statistic

    Parameters
    ----------
    target: array of shape(nt,anat_dim) or None
            the target positions
    data: array of shape(nd,anat_dim) or None
          the data position
    sigma=1.0 (float), kernel parameter
              or  a distance that say how good good is

    Returns
    -------
    sensitivity (float): how well the targets are fitted
                by the data  in [0,1] interval
                1 is good
                0 is bad
    """
    from ...algorithms.utils.fast_distance import euclidean_distance as ed
    if data == None:
        if target == None:
            return 0.# could be 1.0 ?
        else:
            return 0.
    if target == None:
        return 0.

    dmatrix = ed(data, target) / sigma
    sensitivity = dmatrix.min(0)
    sensitivity = np.exp( - 0.5 * sensitivity ** 2)
    sensitivity = np.mean(sensitivity)
    return sensitivity


# -------------------------------------------------------
# ---------- The main functions -----------------------------
# -------------------------------------------------------


def voxel_reproducibility(data, vardata, domain, ngroups, method='crfx',
                          swap=False, verbose=0, **kwargs):
    """ return a measure of voxel-level reproducibility of activation patterns

    Parameters
    ----------
    data: array of shape (nvox,nsubj)
          the input data from which everything is computed
    vardata: array of shape (nvox,nsubj)
             the corresponding variance information
             ngroups (int):
             Number of subbgroups to be drawn
    domain: referential- and domain-defining image
    ngourps: int,
             number of groups to be used in the resampling procedure
    method: string, to be chosen among 'crfx', 'cmfx', 'cffx'
            inference method under study
    verbose: bool, verbosity mode

    Returns
    -------
    kappa (float): the desired  reproducibility index
    """
    rmap = map_reproducibility(data, vardata, domain, ngroups, method,
                                     swap, verbose, **kwargs)

    h = np.array([np.sum(rmap == i) for i in range(ngroups + 1)])
    hr = histo_repro(h)
    return hr


def draw_samples(nsubj, ngroups, split_method='default'):
    """ Draw randomly ngroups sets of samples from [0..nsubj-1]

    Parameters
    ----------
    nsubj, int, the total number of items
    ngroups, int, the number of desired groups
    split_method: string, optional,
                  to be chosen among 'default', 'bootstrap', 'jacknife'
                  if 'bootstrap', then each group will be nsubj
                     drawn with repetitions among nsubj
                  if 'jacknife' the population is divided into
                      ngroups disjoint equally-sized subgroups
                  if 'default', 'bootstrap' is used when nsubj < 10 * ngroups
                     otherwise jacknife is used

    Returns
    -------
    samples, a list of ngroups array that represent the subsets.

    fixme : this should allow variable bootstrap,
    i.e. draw ngroups of groupsize among nsubj
    """
    if split_method == 'default':
        if nsubj > 10 * ngroups:
            samples = split_group(nsubj, ngroups)
        else:
            samples = bootstrap_group(nsubj, ngroups)
    elif split_method == 'bootstrap':
        samples = bootstrap_group(nsubj, ngroups)
    elif split_method == '':
        samples = split_group(nsubj, ngroups)
    else:
        raise ValueError('unknown splitting method')

    return samples


def map_reproducibility(data, vardata, domain, ngroups, method='crfx',
                        swap=False, verbose=0, **kwargs):
    """ Return a reproducibility map for the given method

    Parameters
    ----------
    data: array of shape (nvox,nsubj)
          the input data from which everything is computed
    vardata: array of the same size
             the corresponding variance information
    domain: referential- and domain-defining image
    ngroups (int): the size of each subrgoup to be studied
    threshold (float): binarization threshold
              (makes sense only if method==rfx)
    method='crfx', string to be chosen among 'crfx', 'cmfx', 'cffx'
           inference method under study
    verbose=0 : verbosity mode

    Returns
    -------
    rmap: array of shape(nvox)
          the reproducibility map
    """
    nsubj = data.shape[1]
    nvox = data.shape[0]
    samples = draw_samples(nsubj, ngroups)
    rmap = np.zeros(nvox)

    for i in range(ngroups):
        x = data[:, samples[i]]

        if swap:
            # randomly swap the sign of x
            x *= (2 * (np.random.rand(len(samples[i])) > 0.5) - 1)

        if method is not 'crfx':
            vx = vardata[:, samples[i]]
        csize = kwargs['csize']
        threshold = kwargs['threshold']

        # compute the statistical maps according to the method you like
        if method == 'crfx':
            stat_map = ttest(x)
        elif method == 'cffx':
            stat_map = fttest(x, vx)
        elif method == 'cmfx':
            stat_map = mfx_ttest(x, vx)
        elif method == 'cjt':
            # if kwargs.has_key('k'):
            if 'k' in kwargs:
                k = kwargs['k']
            else:
                k = nsubj / 2
            stat_map = conjunction(x, vx, k)
        else:
            raise ValueError('unknown method')

        # add the binarized map to a reproducibility map
        rmap += cluster_threshold(stat_map, domain, threshold, csize) > 0

    return rmap


def peak_reproducibility(data, vardata, domain, ngroups, sigma, method='crfx',
                         swap=False, verbose=0, **kwargs):
    """ Return a measure of cluster-level reproducibility
    of activation patterns
    (i.e. how far clusters are from each other)

    Parameters
    ----------
    data: array of shape (nvox,nsubj)
          the input data from which everything is computed
    vardata: array of shape (nvox,nsubj)
             the variance of the data that is also available
    domain: refenrtial- and domain-defining image
    ngroups (int),
             Number of subbgroups to be drawn
    sigma: float, parameter that encodes how far far is
    threshold: float, binarization threshold
    method: string to be chosen among 'crfx', 'cmfx' or 'cffx',
           inference method under study
    swap = False: if True, a random sign swap of the data is performed
         This is used to simulate a null hypothesis on the data.
    verbose=0 : verbosity mode

    Returns
    -------
    score (float): the desired  cluster-level reproducibility index
    """
    tiny = 1.e-15
    nsubj = data.shape[1]
    samples = draw_samples(nsubj, ngroups)
    all_pos = []

    # compute the positions in the different subgroups
    for i in range(ngroups):
        x = data[:, samples[i]]

        if swap:
            # apply a random sign swap to x
            x *= (2 * (np.random.rand(len(samples[i])) > 0.5) - 1)

        if method is not 'crfx':
            vx = vardata[:, samples[i]]
        if method is not 'bsa':
            threshold = kwargs['threshold']

            if method == 'crfx':
                stat_map = ttest(x)
            elif method == 'cmfx':
                stat_map = mfx_ttest(x, vx)
            elif method == 'cffx':
                stat_map = fttest(x, vx)
            elif method == 'cjt':
                if 'k' in kwargs:
                    k = kwargs['k']
                else:
                    k = nsubj / 2
                stat_map = conjunction(x, vx, k)

            pos = get_peak_position_from_thresholded_map(
                stat_map, domain, threshold)
            all_pos.append(pos)
        else:
            # method='bsa' is a special case
            tx = x / (tiny + np.sqrt(vx))
            afname = kwargs['afname']
            theta = kwargs['theta']
            dmax = kwargs['dmax']
            ths = kwargs['ths']
            thq = kwargs['thq']
            smin = kwargs['smin']
            niter = kwargs['niter']
            afname = afname + '_%02d_%04d.pic' % (niter, i)
            pos = coord_bsa(domain, tx, theta, dmax, ths, thq, smin, afname)
            all_pos.append(pos)

    # derive a kernel-based goodness measure from the pairwise comparison
    # of sets of positions
    score = 0
    for i in range(ngroups):
        for j in range(i):
            score += statistics_from_position(all_pos[i], all_pos[j], sigma)
            score += statistics_from_position(all_pos[j], all_pos[i], sigma)
    score /= (ngroups * (ngroups - 1))
    return score


def cluster_reproducibility(data, vardata, domain, ngroups, sigma,
                            method='crfx', swap=False, verbose=0,
                            **kwargs):
    """Returns a measure of cluster-level reproducibility
    of activation patterns
    (i.e. how far clusters are from each other)

    Parameters
    ----------
    data: array of shape (nvox,nsubj)
          the input data from which everything is computed
    vardata: array of shape (nvox,nsubj)
             the variance of the data that is also available
    domain: referential- and domain- defining image instance
    ngroups (int),
             Number of subbgroups to be drawn
    sigma (float): parameter that encodes how far far is
    threshold (float):
              binarization threshold
    method='crfx', string to be chosen among 'crfx', 'cmfx' or 'cffx'
           inference method under study
    swap = False: if True, a random sign swap of the data is performed
         This is used to simulate a null hypothesis on the data.
    verbose=0 : verbosity mode

    Returns
    -------
    score (float): the desired  cluster-level reproducibility index
    """
    tiny = 1.e-15
    nsubj = data.shape[1]
    samples = draw_samples(nsubj, ngroups)
    all_pos = []

    # compute the positions in the different subgroups
    for i in range(ngroups):
        x = data[:, samples[i]]

        if swap:
            # apply a random sign swap to x
            x *= (2 * (np.random.rand(len(samples[i])) > 0.5) - 1)

        if method is not 'crfx':
            vx = vardata[:, samples[i]]
        if method is not 'bsa':
            csize = kwargs['csize']
            threshold = kwargs['threshold']
            if method == 'crfx':
                stat_map = ttest(x)
            elif method == 'cmfx':
                stat_map = mfx_ttest(x, vx)
            elif method == 'cffx':
                stat_map = fttest(x, vx)
            elif method == 'cjt':
                if  'k' in kwargs:
                    k = kwargs['k']
                else:
                    k = nsubj / 2
                stat_map = conjunction(x, vx, k)
            pos = get_cluster_position_from_thresholded_map(stat_map, domain,
                                                            threshold, csize)
            all_pos.append(pos)
        else:
            # method='bsa' is a special case
            tx = x / (tiny + np.sqrt(vx))
            afname = kwargs['afname']
            theta = kwargs['theta']
            dmax = kwargs['dmax']
            ths = kwargs['ths']
            thq = kwargs['thq']
            smin = kwargs['smin']
            niter = kwargs['niter']
            afname = afname + '_%02d_%04d.pic' % (niter, i)
            pos = coord_bsa(domain, tx, theta, dmax, ths, thq, smin, afname)
        all_pos.append(pos)

    # derive a kernel-based goodness measure from the pairwise comparison
    # of sets of positions
    score = 0
    for i in range(ngroups):
        for j in range(i):
            score += statistics_from_position(all_pos[i], all_pos[j], sigma)
            score += statistics_from_position(all_pos[j], all_pos[i], sigma)

    score /= (ngroups * (ngroups - 1))
    return score


def group_reproducibility_metrics(
    mask_images, contrast_images, variance_images, thresholds, ngroups,
    method, cluster_threshold=10, number_of_samples=10, sigma=6.,
    do_clusters=True, do_voxels=True, do_peaks=True, swap=False):
    """
    Main function to perform reproducibility analysis, including nifti1 io

    Parameters
    ----------
    threshold: list or 1-d array,
               the thresholds to be tested

    Returns
    -------
    cluster_rep_results: dictionary,
                         results of cluster-level reproducibility analysi
    voxel_rep_results: dictionary,
                       results of voxel-level reproducibility analysis
    peak_rep_results: dictionary,
                      results of peak-level reproducibility analysis
    """
    from nibabel import load
    from ..mask import intersect_masks

    if ((len(variance_images) == 0) & (method is not 'crfx')):
        raise ValueError('Variance images are necessary')

    nsubj = len(contrast_images)

    # compute the group mask
    affine = load(mask_images[0]).get_affine()
    mask = intersect_masks(mask_images, threshold=0) > 0
    domain = grid_domain_from_binary_array(mask, affine)

    # read the data
    group_con = []
    group_var = []
    for s in range(nsubj):
        group_con.append(load(contrast_images[s]).get_data()[mask])
        if len(variance_images) > 0:
            group_var.append(load(variance_images[s]).get_data()[mask])

    group_con = np.squeeze(np.array(group_con)).T
    group_con[np.isnan(group_con)] = 0
    if len(variance_images) > 0:
        group_var = np.squeeze(np.array(group_var)).T
        group_var[np.isnan(group_var)] = 0
        group_var = np.maximum(group_var, 1.e-15)

    # perform the analysis
    voxel_rep_results = {}
    cluster_rep_results = {}
    peak_rep_results = {}

    for ng in ngroups:
        if do_voxels:
            voxel_rep_results.update({ng: {}})
        if do_clusters:
            cluster_rep_results.update({ng: {}})
        if do_peaks:
            peak_rep_results.update({ng: {}})
        for th in thresholds:
            kappa = []
            cls = []
            pk = []
            kwargs = {'threshold': th, 'csize': cluster_threshold}

            for i in range(number_of_samples):
                if do_voxels:
                    kappa.append(voxel_reproducibility(
                            group_con, group_var, domain, ng, method, swap,
                            **kwargs))
                if do_clusters:
                    cls.append(cluster_reproducibility(
                            group_con, group_var, domain, ng, sigma, method,
                            swap, **kwargs))
                if do_peaks:
                    pk.append(peak_reproducibility(
                            group_con, group_var, domain, ng, sigma, method,
                            swap, **kwargs))

            if do_voxels:
                voxel_rep_results[ng].update({th: np.array(kappa)})
            if do_clusters:
                cluster_rep_results[ng].update({th: np.array(cls)})
            if do_peaks:
                peak_rep_results[ng].update({th: np.array(cls)})

    return voxel_rep_results, cluster_rep_results, peak_rep_results


# -------------------------------------------------------
# ---------- BSA stuff ----------------------------------
# -------------------------------------------------------


def coord_bsa(domain, betas, theta=3., dmax=5., ths=0, thq=0.5, smin=0,
              afname=None):
    """ main function for  performing bsa on a dataset
    where bsa =  nipy.labs.spatial_models.bayesian_structural_analysis

    Parameters
    ----------
    domain: image instance,
          referential- and domain-defining image
    betas: array of shape (nbnodes, subjects),
           the multi-subject statistical maps
    theta: float, optional
           first level threshold
    dmax: float>0, optional
          expected cluster std in the common space in units of coord
    ths: int, >=0), optional
         representatitivity threshold
    thq: float, optional,
         posterior significance threshold should be in [0,1]
    smin: int, optional,
          minimal size of the regions to validate them
    afname: string, optional
            path where intermediate resullts cam be pickelized

    Returns
    -------
    afcoord array of shape(number_of_regions,3):
            coordinate of the found landmark regions
    """
    from ..spatial_models.bayesian_structural_analysis import compute_BSA_quick

    crmap, AF, BF, p = compute_BSA_quick(
        domain, betas, dmax, thq, smin, ths, theta, verbose=0)
    if AF == None:
        return None
    if afname is not None:
        import pickle
        pickle.dump(AF, afname)
    afcoord = AF.discrete_to_roi_features('position')
    return afcoord



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Utilities for working with matrices """

import numpy as np
import scipy.linalg as spl

def matrix_rank(M, tol=None):
    ''' Return rank of matrix using SVD method

    Rank of the array is the number of SVD singular values of the
    array that are greater than `tol`.

    This version of matrix rank is very similar to the numpy.linalg version
    except for the use of:

    * scipy.linalg.svd istead of numpy.linalg.svd.
    * the MATLAB algorithm for default tolerance calculation

    ``matrix_rank`` appeared in numpy.linalg in December 2009, first available
    in numpy 1.5.0.

    Parameters
    ----------
    M : array-like
        array of <=2 dimensions
    tol : {None, float}
         threshold below which SVD values are considered zero. If `tol`
         is None, and `S` is an array with singular values for `M`, and
         `eps` is the epsilon value for datatype of `S`, then `tol` set
         to ``S.max() * eps * max(M.shape)``.

    Examples
    --------
    >>> matrix_rank(np.eye(4)) # Full rank matrix
    4
    >>> I=np.eye(4); I[-1,-1] = 0. # rank deficient matrix
    >>> matrix_rank(I)
    3
    >>> matrix_rank(np.zeros((4,4))) # All zeros - zero rank
    0
    >>> matrix_rank(np.ones((4,))) # 1 dimension - rank 1 unless all 0
    1
    >>> matrix_rank(np.zeros((4,)))
    0
    >>> matrix_rank([1]) # accepts array-like
    1

    Notes
    -----
    We check for numerical rank deficiency by using tol=max(M.shape) * eps *
    S[0] (where S[0] is the maximum singular value and thus the 2-norm of the
    matrix). This is one tolerance threshold for rank deficiency, and the
    default algorithm used by MATLAB [2].  When floating point roundoff is the
    main concern, then "numerical rank deficiency" is a reasonable choice. In
    some cases you may prefer other definitions. The most useful measure of the
    tolerance depends on the operations you intend to use on your matrix. For
    example, if your data come from uncertain measurements with uncertainties
    greater than floating point epsilon, choosing a tolerance near that
    uncertainty may be preferable.  The tolerance may be absolute if the
    uncertainties are absolute rather than relative.

    References
    ----------
    .. [1] G. H. Golub and C. F. Van Loan, _Matrix Computations_.
    Baltimore: Johns Hopkins University Press, 1996.
    .. [2] http://www.mathworks.com/help/techdoc/ref/rank.html
    '''
    M = np.asarray(M)
    if M.ndim > 2:
        raise TypeError('array should have 2 or fewer dimensions')
    if M.ndim < 2:
        return int(not np.all(M==0))
    S = spl.svd(M, compute_uv=False)
    if tol is None:
        tol = S.max() * np.finfo(S.dtype).eps * max(M.shape)
    return np.sum(S > tol)


def full_rank(X, r=None):
    """ Return full-rank matrix whose column span is the same as X

    Uses an SVD decomposition.

    If the rank of `X` is known it can be specified by `r` -- no check is made
    to ensure that this really is the rank of X.

    Parameters
    ----------
    X : array-like
        2D array which may not be of full rank.
    r : None or int
        Known rank of `X`.  r=None results in standard matrix rank calculation.
        We do not check `r` is really the rank of X; it is to speed up
        calculations when the rank is already known.

    Returns
    -------
    fX : array
        Full-rank matrix with column span matching that of `X`
    """
    if r is None:
        r = matrix_rank(X)
    V, D, U = spl.svd(X, full_matrices=0)
    order = np.argsort(D)
    order = order[::-1]
    value = []
    for i in range(r):
        value.append(V[:,order[i]])
    return np.asarray(np.transpose(value)).astype(np.float64)


def pos_recipr(X):
    """ Return element-wise reciprocal of array, setting `X`>=0 to 0

    Return the reciprocal of an array, setting all entries less than or
    equal to 0 to 0. Therefore, it presumes that X should be positive in
    general.

    Parameters
    ----------
    X : array-like

    Returns
    -------
    rX : array
       array of same shape as `X`, dtype np.float, with values set to
       1/X where X > 0, 0 otherwise
    """
    X = np.asarray(X)
    return np.where(X<=0, 0, 1. / X)


def recipr0(X):
    """ Return element-wise reciprocal of array, `X`==0 -> 0

    Return the reciprocal of an array, setting all entries equal to 0 as 0. It
    does not assume that X should be positive in general.

    Parameters
    ----------
    X : array-like

    Returns
    -------
    rX : array
    """
    X = np.asarray(X)
    return np.where(X==0, 0, 1. / X)





# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import matplotlib.pyplot as plt
plt.plot([1,2,3], [4,5,6])
plt.ylabel('some more numbers')




""" Testing coordinate map defined spaces
"""

import numpy as np

from nibabel.affines import from_matvec

from ...image.image import Image
from ..coordinate_system import CoordinateSystem as CS, CoordSysMakerError
from ..coordinate_map import AffineTransform, CoordinateMap
from ..spaces import (vox2mni, vox2scanner, vox2talairach, vox2unknown,
                      vox2aligned, xyz_affine, xyz_order, SpaceTypeError,
                      AxesError, AffineError, XYZSpace, known_space,
                      known_spaces, is_xyz_space, SpaceError,
                      is_xyz_affable,
                      get_world_cs, mni_csm, mni_space)

from numpy.testing import (assert_array_almost_equal,
                           assert_array_equal)

from nose.tools import (assert_true, assert_false, assert_equal, assert_raises,
                        assert_not_equal)

VARS = {}


def setup():
    d_names = list('ijkl')
    xyzs = 'x=L->R', 'y=P->A', 'z=I->S'
    mni_xyzs = ['mni-' + suff for suff in xyzs]
    scanner_xyzs = ['scanner-' + suff for suff in xyzs]
    unknown_xyzs = ['unknown-' + suff for suff in xyzs]
    aligned_xyzs = ['aligned-' + suff for suff in xyzs]
    talairach_xyzs = ['talairach-' + suff for suff in xyzs]
    r_names = mni_xyzs + ['t']
    d_cs_r3 = CS(d_names[:3], 'voxels')
    d_cs_r4 = CS(d_names[:4], 'voxels')
    r_cs_r3 = CS(r_names[:3], 'mni')
    r_cs_r4 = CS(r_names[:4], 'mni')
    VARS.update(locals())


def test_xyz_space():
    # Space objects
    sp = XYZSpace('hijo')
    assert_equal(sp.name, 'hijo')
    exp_labels = ['hijo-' + L for L in 'x=L->R', 'y=P->A', 'z=I->S']
    exp_map = dict(zip('xyz', exp_labels))
    assert_equal([sp.x, sp.y, sp.z], exp_labels)
    assert_equal(sp.as_tuple(), tuple(exp_labels))
    assert_equal(sp.as_map(), exp_map)
    known = {}
    sp.register_to(known)
    assert_equal(known, dict(zip(exp_labels, 'xyz')))
    # Coordinate system making, and __contains__ tests
    csm = sp.to_coordsys_maker()
    cs = csm(2)
    assert_equal(cs, CS(exp_labels[:2], 'hijo'))
    # This is only 2 dimensions, not fully in space
    assert_false(cs in sp)
    cs = csm(3)
    assert_equal(cs, CS(exp_labels, 'hijo'))
    # We now have all 3, this in in the space
    assert_true(cs in sp)
    # More dimensions than default, error
    assert_raises(CoordSysMakerError, csm, 4)
    # But we can pass in names for further dimensions
    csm = sp.to_coordsys_maker('tuv')
    cs = csm(6)
    assert_equal(cs, CS(exp_labels + list('tuv'), 'hijo'))
    # These are also in the space, because they contain xyz
    assert_true(cs in sp)
    # The axes can be in any order as long as they are a subset
    cs = CS(exp_labels, 'hijo')
    assert_true(cs in sp)
    cs = CS(exp_labels[::-1], 'hijo')
    assert_true(cs in sp)
    cs = CS(['t'] + exp_labels, 'hijo')
    assert_true(cs in sp)
    # The coordinate system name doesn't matter
    cs = CS(exp_labels, 'hija')
    assert_true(cs in sp)
    # Images, and coordinate maps, also work
    cmap = AffineTransform('ijk', cs, np.eye(4))
    assert_true(cmap in sp)
    img = Image(np.zeros((2,3,4)), cmap)
    assert_true(img in sp)
    # equality
    assert_equal(XYZSpace('hijo'), XYZSpace('hijo'))
    assert_not_equal(XYZSpace('hijo'), XYZSpace('hija'))


def test_is_xyz_space():
    # test test for xyz space
    assert_true(is_xyz_space(XYZSpace('hijo')))
    for sp in known_spaces:
        assert_true(is_xyz_space(sp))
    for obj in ([], {}, object(), CS('xyz')):
        assert_false(is_xyz_space(obj))


def test_known_space():
    # Known space utility routine
    for sp in known_spaces:
        cs = sp.to_coordsys_maker()(3)
        assert_equal(known_space(cs), sp)
    cs = CS('xyz')
    assert_equal(known_space(cs), None)
    sp0 = XYZSpace('hijo')
    sp1 = XYZSpace('hija')
    custom_spaces = (sp0, sp1)
    for sp in custom_spaces:
        cs = sp.to_coordsys_maker()(3)
        assert_equal(known_space(cs, custom_spaces), sp)


def test_image_creation():
    # 3D image
    arr = np.arange(24).reshape(2,3,4)
    aff = np.diag([2,3,4,1])
    img = Image(arr, vox2mni(aff))
    assert_equal(img.shape, (2,3,4))
    assert_array_equal(img.affine, aff)
    assert_array_equal(img.coordmap,
                       AffineTransform(VARS['d_cs_r3'], VARS['r_cs_r3'], aff))
    # 4D image
    arr = np.arange(24).reshape(2,3,4,1)
    img = Image(arr, vox2mni(aff, 7))
    exp_aff = np.diag([2,3,4,7,1])
    assert_equal(img.shape, (2,3,4,1))
    exp_cmap = AffineTransform(VARS['d_cs_r4'], VARS['r_cs_r4'], exp_aff)
    assert_equal(img.coordmap, exp_cmap)


def test_default_makers():
    # Tests that the makers make expected coordinate maps
    for csm, r_names, r_name in (
        (vox2scanner, VARS['scanner_xyzs'] + ['t'], 'scanner'),
        (vox2unknown, VARS['unknown_xyzs'] + ['t'], 'unknown'),
        (vox2aligned, VARS['aligned_xyzs'] + ['t'], 'aligned'),
        (vox2mni, VARS['mni_xyzs'] + ['t'], 'mni'),
        (vox2talairach, VARS['talairach_xyzs'] + ['t'], 'talairach')):
        for i in range(1,5):
            dom_cs = CS('ijkl'[:i], 'voxels')
            ran_cs = CS(r_names[:i], r_name)
            aff = np.diag(range(i) + [1])
            assert_equal(csm(aff), AffineTransform(dom_cs, ran_cs, aff))


def test_get_world_cs():
    # Utility to get world from a variety of inputs
    assert_equal(get_world_cs('mni'), mni_csm(3))
    mnit = mni_space.to_coordsys_maker('t')(4)
    assert_equal(get_world_cs(mni_space, 4), mnit)
    assert_equal(get_world_cs(mni_csm, 4), mni_csm(4))
    assert_equal(get_world_cs(CS('xyz')), CS('xyz'))
    hija = XYZSpace('hija')
    maker = hija.to_coordsys_maker('qrs')
    assert_equal(get_world_cs('hija', ndim = 5, extras='qrs', spaces=[hija]),
                 maker(5))
    assert_raises(SpaceError, get_world_cs, 'hijo')
    assert_raises(SpaceError, get_world_cs, 'hijo', spaces=[hija])
    assert_raises(ValueError, get_world_cs, 0)


def test_xyz_affine():
    # Getting an xyz affine from coordmaps
    aff3d = from_matvec(np.arange(9).reshape((3,3)), [15,16,17])
    cmap3d = AffineTransform(VARS['d_cs_r3'], VARS['r_cs_r3'], aff3d)
    rzs = np.c_[np.arange(12).reshape((4,3)), [0,0,0,12]]
    aff4d = from_matvec(rzs, [15,16,17,18])
    cmap4d = AffineTransform(VARS['d_cs_r4'], VARS['r_cs_r4'], aff4d)
    # Simplest case of 3D affine -> affine unchanged
    assert_array_equal(xyz_affine(cmap3d), aff3d)
    # 4D (5, 5) affine -> 3D equivalent
    assert_array_equal(xyz_affine(cmap4d), aff3d)
    # Any dimensions not spatial, AxesError
    r_cs = CS(('mni-x', 'mni-y', 'mni-q'), 'mni')
    funny_cmap = AffineTransform(VARS['d_cs_r3'],r_cs, aff3d)
    assert_raises(AxesError, xyz_affine, funny_cmap)
    r_cs = CS(('mni-x', 'mni-q', 'mni-z'), 'mni')
    funny_cmap = AffineTransform(VARS['d_cs_r3'],r_cs, aff3d)
    assert_raises(AxesError, xyz_affine, funny_cmap)
    # We insist that the coordmap is in output xyz order
    permutations = (0, 2, 1), (1, 0, 2), (1, 2, 0), (2, 0, 1), (2, 1, 0)
    for perm in permutations:
        assert_raises(AxesError, xyz_affine, cmap3d.reordered_range(perm))
    # The input order doesn't matter, as long as the xyz axes map to the first
    # three input axes
    for perm in permutations:
        assert_array_equal(xyz_affine(
            cmap3d.reordered_domain(perm)), aff3d[:, perm + (-1,)])
    # But if the corresponding input axes not in the first three, an axis error
    wrong_inputs = cmap4d.reordered_domain([0, 1, 3, 2])
    assert_raises(AxesError, xyz_affine, wrong_inputs)
    # xyzs must be orthogonal to dropped axis
    for i in range(3):
        aff = aff4d.copy()
        aff[i,3] = 1
        cmap = AffineTransform(VARS['d_cs_r4'], VARS['r_cs_r4'], aff)
        assert_raises(AffineError, xyz_affine, cmap)
        # And if reordered
        assert_raises(AxesError, xyz_affine, cmap.reordered_range([2,0,1,3]))
    # Non-square goes to square
    aff54 = np.array([[0, 1, 2, 15],
                      [3, 4, 5, 16],
                      [6, 7, 8, 17],
                      [0, 0, 0, 18],
                      [0, 0, 0, 1]])
    cmap = AffineTransform(VARS['d_cs_r3'], VARS['r_cs_r4'], aff54)
    assert_array_equal(xyz_affine(cmap), aff3d)
    aff57 = np.array([[0, 1, 2, 0, 0, 0, 15],
                      [3, 4, 5, 0, 0, 0, 16],
                      [6, 7, 8, 0, 0, 0, 17],
                      [0, 0, 0, 0, 0, 0, 18],
                      [0, 0, 0, 0, 0, 0, 1]])
    d_cs_r6 = CS('ijklmn', 'voxels')
    cmap = AffineTransform(d_cs_r6, VARS['r_cs_r4'], aff57)
    assert_array_equal(xyz_affine(cmap), aff3d)
    # Non-affine raises SpaceTypeError
    cmap_cmap = CoordinateMap(VARS['d_cs_r4'], VARS['r_cs_r4'], lambda x:x*3)
    assert_raises(SpaceTypeError, xyz_affine, cmap_cmap)
    # Not enough dimensions - SpaceTypeError
    d_cs_r2 = CS('ij', 'voxels')
    r_cs_r2 = CS(VARS['r_names'][:2], 'mni')
    cmap = AffineTransform(d_cs_r2, r_cs_r2,
                           np.array([[2,0,10],[0,3,11],[0,0,1]]))
    assert_raises(AxesError, xyz_affine, cmap)
    # Can pass in own validator
    my_valtor = dict(blind='x', leading='y', ditch='z')
    r_cs = CS(('blind', 'leading', 'ditch'), 'fall')
    cmap = AffineTransform(VARS['d_cs_r3'],r_cs, aff3d)
    assert_raises(AxesError, xyz_affine, cmap)
    assert_array_equal(xyz_affine(cmap, my_valtor), aff3d)
    # Slices in x, y, z coordmaps raise error because of missing spatial
    # dimensions
    arr = np.arange(120).reshape((2, 3, 4, 5))
    aff = np.diag([2, 3, 4, 5, 1])
    img = Image(arr, vox2mni(aff))
    assert_raises(AxesError, xyz_affine, img[1].coordmap)
    assert_raises(AxesError, xyz_affine, img[:,1].coordmap)
    assert_raises(AxesError, xyz_affine, img[:,:,1].coordmap)


def test_xyz_order():
    # Getting xyz ordering from a coordinate system
    assert_array_equal(xyz_order(VARS['r_cs_r3']), [0,1,2])
    assert_array_equal(xyz_order(VARS['r_cs_r4']), [0,1,2,3])
    r_cs = CS(('mni-x=L->R', 'mni-y=P->A', 'mni-q'), 'mni')
    assert_raises(AxesError, xyz_order, r_cs)
    r_cs = CS(('t', 'mni-x=L->R', 'mni-z=I->S', 'mni-y=P->A'), 'mni')
    assert_array_equal(xyz_order(r_cs), [1, 3, 2, 0])
    # Can pass in own validator
    my_valtor = dict(ditch='x', leading='y', blind='z')
    r_cs = CS(('blind', 'leading', 'ditch'), 'fall')
    assert_raises(AxesError, xyz_order, r_cs)
    assert_array_equal(xyz_order(r_cs, my_valtor), [2,1,0])


def test_is_xyz_affable():
    # Whether there exists an xyz affine for this coordmap
    affine = np.diag([2,4,5,6,1])
    cmap = AffineTransform(VARS['d_cs_r4'], VARS['r_cs_r4'], affine)
    assert_true(is_xyz_affable(cmap))
    assert_false(is_xyz_affable(cmap.reordered_range([3,0,1,2])))
    assert_false(is_xyz_affable(cmap.reordered_domain([3,0,1,2])))
    # Can pass in own validator
    my_valtor = dict(blind='x', leading='y', ditch='z')
    r_cs = CS(('blind', 'leading', 'ditch'), 'fall')
    affine = from_matvec(np.arange(9).reshape((3, 3)), [11, 12, 13])
    cmap = AffineTransform(VARS['d_cs_r3'], r_cs, affine)
    # No xyz affine if we don't use our custom dictionary
    assert_false(is_xyz_affable(cmap))
    # Is if we do
    assert_true(is_xyz_affable(cmap, my_valtor))



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Script to auto-generate our API docs.

This script should run in Python 2 and Python 3
"""
# stdlib imports
import os

# local imports
from apigen import ApiDocWriter

#*****************************************************************************
if __name__ == '__main__':
    package = 'nipy'
    outdir = os.path.join('api','generated')
    docwriter = ApiDocWriter(package)
    docwriter.package_skip_patterns += [r'\.fixes$',
                                        r'\.externals$',
                                        #r'\.labs\.viz',
                                        ]
    # XXX: Avoid nipy.modalities.fmri.aliased due to a bug in python2.6
    docwriter.module_skip_patterns += [r'\.modalities\.fmri.aliased',
                                        ]
    docwriter.write_api_docs(outdir)
    docwriter.write_index(outdir, 'gen', relative_to='api')
    print('%d files written' % len(docwriter.written_modules))



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function
__doc__ = \
"""
This example shows how to get variance and beta estimated from a nipy GLM.

More specifically:

1. A sequence of fMRI volumes are loaded.
2. A design matrix describing all the effects related to the data is computed.
3. A GLM is applied to the dataset, effect and variance images are produced.

Note that this corresponds to a single run.

Needs matplotlib

Author : Bertrand Thirion, 2010--2012
"""
print(__doc__)

from os import mkdir, getcwd, path

import numpy as np

try:
    import matplotlib.pyplot as plt
except ImportError:
    raise RuntimeError("This script needs the matplotlib library")

from nibabel import Nifti1Image, save

from nipy.modalities.fmri.glm import FMRILinearModel
from nipy.modalities.fmri.design_matrix import make_dmtx
from nipy.modalities.fmri.experimental_paradigm import \
    load_paradigm_from_csv_file
from nipy.labs.viz import plot_map, cm

# Local import
from get_data_light import DATA_DIR, get_first_level_dataset

#######################################
# Data and analysis parameters
#######################################

# volume mask
# This dataset is large
get_first_level_dataset()
data_path = path.join(DATA_DIR, 's12069_swaloc1_corr.nii.gz')
paradigm_file = path.join(DATA_DIR, 'localizer_paradigm.csv')

# timing
n_scans = 128
tr = 2.4

# paradigm
frametimes = np.linspace(0, (n_scans - 1) * tr, n_scans)

# confounds
hrf_model = 'canonical'
drift_model = "cosine"
hfcut = 128

# write directory
write_dir = path.join(getcwd(), 'results')
if not path.exists(write_dir):
    mkdir(write_dir)

print('Computation will be performed in directory: %s' % write_dir)

########################################
# Design matrix
########################################

print('Loading design matrix...')

# the example example.labs.write_paradigm_file shows how to create this file
paradigm = load_paradigm_from_csv_file(paradigm_file)['0']

design_matrix = make_dmtx(frametimes, paradigm, hrf_model=hrf_model,
                          drift_model=drift_model, hfcut=hfcut)

ax = design_matrix.show()
ax.set_position([.05, .25, .9, .65])
ax.set_title('Design matrix')

plt.savefig(path.join(write_dir, 'design_matrix.png'))
dim = design_matrix.matrix.shape[1]

########################################
# Perform a GLM analysis
########################################

print('Fitting a GLM (this takes time)...')
fmri_glm = FMRILinearModel(data_path, design_matrix.matrix,
                           mask='compute')
fmri_glm.fit(do_scaling=True, model='ar1')

########################################
# Output beta and variance images
########################################
beta_hat = fmri_glm.glms[0].get_beta()  # Least-squares estimates of the beta
variance_hat = fmri_glm.glms[0].get_mse() # Estimates of the variance
mask = fmri_glm.mask.get_data() > 0

# output beta images
beta_map = np.tile(mask.astype(np.float)[..., np.newaxis], dim)
beta_map[mask] = beta_hat.T
beta_image = Nifti1Image(beta_map, fmri_glm.affine)
beta_image.get_header()['descrip'] = (
    'Parameter estimates of the localizer dataset')
save(beta_image, path.join(write_dir, 'beta.nii'))
print("Beta image witten in %s" % write_dir)

variance_map = mask.astype(np.float)
variance_map[mask] = variance_hat

# Create a snapshots of the variance image contrasts
vmax = np.log(variance_hat.max())
plot_map(np.log(variance_map + .1),
         fmri_glm.affine,
         cmap=cm.hot_black_bone,
         vmin=np.log(0.1),
         vmax=vmax,
         anat=None,
         threshold=.1, alpha=.9)
plt.show()



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Testing fmri utils

"""
import re

import numpy as np

import sympy
from sympy import Symbol, Function, DiracDelta
from nipy.fixes.sympy.utilities.lambdify import lambdify

from nipy.algorithms.statistics.formula import Term

from ..utils import (
    lambdify_t,
    define,
    events,
    blocks,
    interp,
    linear_interp,
    step_function,
    convolve_functions,
    )
from .. import hrf

from nose.tools import (assert_equal, assert_false, raises, assert_raises)

from numpy.testing import (assert_array_equal, assert_array_almost_equal,
                           assert_almost_equal)


t = Term('t')


def test_define():
    expr = sympy.exp(3*t)
    assert_equal(str(expr), 'exp(3*t)')
    newf = define('f', expr)
    assert_equal(str(newf), 'f(t)')
    f = lambdify_t(newf)
    tval = np.random.standard_normal((3,))
    assert_almost_equal(np.exp(3*tval), f(tval))


def test_events():
    # test events utility function
    h = Function('hrf')
    evs = events([3,6,9])
    assert_equal(DiracDelta(-9 + t) + DiracDelta(-6 + t) +
                 DiracDelta(-3 + t), evs)
    evs = events([3,6,9], f=h)
    assert_equal(h(-3 + t) + h(-6 + t) + h(-9 + t), evs)
    # make some beta symbols
    b = [Symbol('b%d' % i, dummy=True) for i in range(3)]
    a = Symbol('a')
    p = b[0] + b[1]*a + b[2]*a**2
    evs = events([3,6,9], amplitudes=[2,1,-1], g=p)
    assert_equal((2*b[1] + 4*b[2] + b[0])*DiracDelta(-3 + t) +
                 (-b[1] + b[0] + b[2])*DiracDelta(-9 + t) +
                 (b[0] + b[1] + b[2])*DiracDelta(-6 + t),
                 evs)
    evs = events([3,6,9], amplitudes=[2,1,-1], g=p, f=h)
    assert_equal((2*b[1] + 4*b[2] + b[0])*h(-3 + t) +
                 (-b[1] + b[0] + b[2])*h(-9 + t) +
                 (b[0] + b[1] + b[2])*h(-6 + t),
                 evs)
    # test no error for numpy int arrays
    onsets = np.array([30, 70, 100], dtype=np.int64)
    evs = events(onsets, f=hrf.glover)


def test_interp():
    times = [0,4,5.]
    values = [2.,4,6]
    for int_func in (interp, linear_interp):
        s = int_func(times, values, np.nan)
        tval = np.array([-0.1,0.1,3.9,4.1,5.1])
        res = lambdify(t, s)(tval)
        assert_array_equal(np.isnan(res),
                           [True, False, False, False, True])
        assert_array_almost_equal(res[1:-1], [2.05, 3.95, 4.2])
        # default is zero fill
        s = int_func(times, values)
        res = lambdify(t, s)(tval)
        assert_array_almost_equal(res, [0, 2.05, 3.95, 4.2, 0])
        # Can be some other value
        s = int_func(times, values, fill=10)
        res = lambdify(t, s)(tval)
        assert_array_almost_equal(res, [10, 2.05, 3.95, 4.2, 10])
        # If fill is None, raises error on interpolation outside bounds
        s = int_func(times, values, fill=None)
        f = lambdify(t, s)
        assert_array_almost_equal(f(tval[1:-1]), [2.05, 3.95, 4.2])
        assert_raises(ValueError, f, tval[:-1])
        # specifying kind as linear is OK
        s = linear_interp(times, values, kind='linear')
        # bounds_check should match fill
        int_func(times, values, bounds_error=False)
        int_func(times, values, fill=None, bounds_error=True)
        assert_raises(ValueError, int_func, times, values, bounds_error=True)
        # fill should match fill value
        int_func(times, values, fill=10, fill_value=10)
        int_func(times, values, fill_value=0)
        assert_raises(ValueError,
                      int_func, times, values, fill=10, fill_value=9)
        int_func(times, values, fill=np.nan, fill_value=np.nan)
        assert_raises(ValueError,
                      int_func, times, values, fill=10, fill_value=np.nan)
        assert_raises(ValueError,
                      int_func, times, values, fill=np.nan, fill_value=0)


@raises(ValueError)
def test_linear_inter_kind():
    linear_interp([0, 1], [1, 2], kind='cubic')


def test_step_function():
    # test step function
    # step function is a function of t
    s = step_function([0,4,5],[2,4,6])
    tval = np.array([-0.1,0,3.9,4,4.1,5.1])
    lam = lambdify(t, s)
    assert_array_equal(lam(tval), [0, 2, 2, 4, 4, 6])
    s = step_function([0,4,5],[4,2,1])
    lam = lambdify(t, s)
    assert_array_equal(lam(tval), [0, 4, 4, 2, 2, 1])
    # Name default
    assert_false(re.match(r'step\d+\(t\)$', str(s)) is None)
    # Name reloaded
    s = step_function([0,4,5],[4,2,1], name='goodie_goodie_yum_yum')
    assert_equal(str(s), 'goodie_goodie_yum_yum(t)')


def test_blocks():
    on_off = [[1,2],[3,4]]
    tval = np.array([0.4,1.4,2.4,3.4])
    b = blocks(on_off)
    lam = lambdify(t, b)
    assert_array_equal(lam(tval), [0, 1, 0, 1])
    b = blocks(on_off, amplitudes=[3,5])
    lam = lambdify(t, b)
    assert_array_equal(lam(tval), [0, 3, 0, 5])
    # Check what happens with names
    # Default is from step function
    assert_false(re.match(r'step\d+\(t\)$', str(b)) is None)
    # Can pass in another
    b = blocks(on_off, name='funky_chicken')
    assert_equal(str(b), 'funky_chicken(t)')





def numerical_convolve(func1, func2, interval, dt):
    mni, mxi = interval
    time = np.arange(mni, mxi, dt)
    vec1 = func1(time).astype(float)
    vec2 = func2(time).astype(float)
    value = np.convolve(vec1, vec2) * dt
    min_s = min(time.size, value.size)
    time = time[:min_s]
    value = value[:min_s]
    return time, value


def test_convolve_functions():
    # replicate convolution
    # This is a square wave on [0,1]
    f1 = (t > 0) * (t < 1)
    # ff1 is the numerical implementation of same
    ff1 = lambdify(t, f1)
    # Time delta
    dt = 1e-3
    # The convolution of ``f1`` with itself is a triangular wave on
    # [0, 2], peaking at 1 with height 1
    tri = convolve_functions(f1, f1, [0, 2], [0, 2], dt, name='conv')
    assert_equal(str(tri), 'conv(t)')
    ftri = lambdify(t, tri)
    time, value = numerical_convolve(ff1, ff1, [0, 2], dt)
    y = ftri(time)
    # numerical convolve about the same as ours
    assert_array_almost_equal(value, y)
    # peak is at 1
    assert_array_almost_equal(time[np.argmax(y)], 1)
    # Flip the interval and get the same result
    for seq1, seq2 in (((0, 2), (2, 0)),
                       ((2, 0), (0, 2)),
                       ((2, 0), (2, 0))):
        tri = convolve_functions(f1, f1, seq1, seq2, dt)
        ftri = lambdify(t, tri)
        y = ftri(time)
        assert_array_almost_equal(value, y)
    # offset square wave by 1 - offset triangle by 1
    f2 = (t > 1) * (t < 2)
    tri = convolve_functions(f1, f2, [0, 3], [0, 3], dt)
    ftri = lambdify(t, tri)
    o1_time = np.arange(0, 3, dt)
    z1s = np.zeros((np.round(1./dt)))
    assert_array_almost_equal(ftri(o1_time), np.r_[z1s, value])
    # Same for input function
    tri = convolve_functions(f2, f1, [0, 3], [0, 3], dt)
    ftri = lambdify(t, tri)
    assert_array_almost_equal(ftri(o1_time), np.r_[z1s, value])
    # 2 seconds for both
    tri = convolve_functions(f2, f2, [0, 4], [0, 4], dt)
    ftri = lambdify(t, tri)
    o2_time = np.arange(0, 4, dt)
    assert_array_almost_equal(ftri(o2_time), np.r_[z1s, z1s, value])
    # offset by -0.5 - offset triangle by -0.5
    f3 = (t > -0.5) * (t < 0.5)
    tri = convolve_functions(f1, f3, [0, 2], [-0.5, 1.5], dt)
    ftri = lambdify(t, tri)
    o1_time = np.arange(-0.5, 1.5, dt)
    assert_array_almost_equal(ftri(o1_time), value)
    # Same for input function
    tri = convolve_functions(f3, f1, [-0.5, 1.5], [0, 2], dt)
    ftri = lambdify(t, tri)
    assert_array_almost_equal(ftri(o1_time), value)
    # -1 second for both
    tri = convolve_functions(f3, f3, [-0.5, 1.5], [-0.5, 1.5], dt)
    ftri = lambdify(t, tri)
    o2_time = np.arange(-1, 1, dt)
    assert_array_almost_equal(ftri(o2_time), value)
    # Check it's OK to be off the dt grid
    tri = convolve_functions(f1, f1, [dt/2, 2 + dt/2], [0, 2], dt, name='conv')
    ftri = lambdify(t, tri)
    assert_array_almost_equal(ftri(time), value, 3)



"""
New generic implementation of multiple regression analysis under noisy
measurements. 
"""

import numpy as np 

_NITER = 2 

TINY = float(np.finfo(np.double).tiny)

def nonzero(x):
    """
    Force strictly positive values. 
    """
    return np.maximum(x, TINY)


def em(Y, VY, X, C=None, niter=_NITER, log_likelihood=False):
    """
    Maximum likelihood regression in a mixed-effect linear model using
    the EM algorithm.

    Parameters
    ----------
    Y : array
        Array of observations. 
    VY : array 
    
    C is the contrast matrix. Conventionally, C is p x q where p
    is the number of regressors. 
    
    OUTPUT: beta, s2
    beta -- array of parameter estimates
    s2 -- array of squared scale parameters.
    
    REFERENCE:
    Keller and Roche, ISBI 2008.
    """
    # Number of observations, regressors and points
    nobs = X.shape[0]
    nreg = X.shape[1]
    npts = np.prod(Y.shape[1:])
    
    # Reshape input array
    y = np.reshape(Y, (nobs, npts))
    vy = np.reshape(VY, (nobs, npts))

    # Compute the projected pseudo-inverse matrix
    if C == None:
        PpX =  np.linalg.pinv(X)
    else:
        A = np.linalg.inv(np.dot(X.T, X)) # (nreg, nreg)
        B = np.linalg.inv(np.dot(np.dot(C.T, A), C)) # (q,q)
        P = np.eye(nreg) - np.dot(np.dot(np.dot(A, C), B), C.T) # (nreg, nreg)
        PpX = np.dot(np.dot(P, A), X.T) # (nreg, nobs)

    # Initialize outputs
    b = np.zeros((nreg, npts))
    yfit = np.zeros((nobs, npts))
    s2 = np.inf

    # EM loop              
    it = 0
    while it < niter: 

        # E-step: posterior mean and variance of each "true" effect
        w1 = 1/nonzero(vy)
        w2 = 1/nonzero(s2)
        vz = 1/(w1+w2)
        z = vz*(w1*y + w2*yfit)

        # M-step: update effect and variance
        b = np.dot(PpX, z) 
        yfit = np.dot(X, b)
        s2 = np.sum((z-yfit)**2 + vz, 0)/float(nobs) 

        # Increase iteration number 
        it += 1

    # Ouput arrays
    B = np.reshape(b, [nreg] + list(Y.shape[1:]))
    S2 = np.reshape(s2, list(Y.shape[1:]))

    # Log-likelihood computation
    if not log_likelihood: 
        return B, S2
    else: 
        return B, S2, _log_likelihood(y, vy, X, b, s2)


def _log_likelihood(y, vy, X, b, s2): 
    res = y - np.dot(X, b) 
    w = nonzero(vy+s2)
    L = np.sum(np.log(w)+ res**2/w) 
    L *= -0.5
    return L



def log_likelihood_ratio(Y, VY, X, C, niter=_NITER):
    """
    Log-likelihood ratio statistic: 2*(log L - log L0) 

    It is asymptotically distributed like a chi-square with rank(C)
    degrees of freedom under the null hypothesis H0: Cb = 0. 
    """

    # Constrained log-likelihood
    B, S2, ll0 = em(Y, VY, X, C, niter, log_likelihood=True)

    # Unconstrained log-likelihood
    B, S2, ll = em(Y, VY, X, None, niter, log_likelihood=True)
    
    # -2 log R = 2*(ll-ll0)
    return np.maximum(2*(ll-ll0), 0.0)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import numpy as np
from numpy.linalg import inv

def ARcovariance(rho, n, cor=False, sigma=1.):
    """
    Return covariance matrix of a sample of length n from an AR(p)
    process with parameters rho.

    INPUTS:

    rho      -- an array of length p
    sigma    -- standard deviation of the white noise
    """

    rho = np.asarray(rho)
    p = rho.shape[0] 
    invK = np.identity(n)
    for i in range(p):
        invK -= np.diag((rho[i] / sigma) * np.ones(n-i-1), k=-i-1)
    K = inv(invK)
    Q = np.dot(K, K.T)
    if cor:
        sd = np.sqrt(np.diag(Q))
        sdout = np.multiply.outer(sd, sd)
        Q /= sd
    return Q

def ARcomponents(rho, n, drho=0.05, cor=False, sigma=1):
    """
    Numerically differentiate covariance matrices
    of AR(p) of length n with respect to AR parameters
    around the value rho.

    If drho is a vector, they are treated as steps in the numerical
    differentiation.
    """

    rho = np.asarray(rho)
    drho = np.asarray(drho)

    p = rho.shape[0]
    value = []

    if drho.shape == ():
        drho = np.ones(p, np.float) * drho
    drho = np.diag(drho)

    Q = ARcovariance(rho, n, cor=cor, sigma=sigma)
    value = [Q]
    for i in range(p):
        value.append((ARcovariance(rho + drho[i], n, cor=cor) - Q) / drho[i,i])
    return np.asarray(value)


if __name__ == "__main__":
    #print np.diag(ARcovariance([0.3], 100, cor=True), k=0)
    print len(ARcomponents([0.321],8, drho=0.02))



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import warnings
import numpy as np

from nibabel.affines import apply_affine

from ...fixes.nibabel import io_orientation

from ...core.image.image_spaces import (make_xyz_image,
                                        xyz_affine,
                                        as_xyz_image)
from .optimizer import configure_optimizer, use_derivatives
from .affine import Rigid
from ._registration import (_cspline_transform,
                            _cspline_sample3d,
                            _cspline_sample4d)


# Module globals
VERBOSE = True  # enables online print statements
SLICE_ORDER = 'ascending'
INTERLEAVED = None
OPTIMIZER = 'ncg'
XTOL = 1e-5
FTOL = 1e-5
GTOL = 1e-5
STEPSIZE = 1e-6
SMALL = 1e-20
MAXITER = 64
MAXFUN = None
BORDERS = 1, 1, 1
REFSCAN = 0
EXTRAPOLATE_SPACE = 'reflect'
EXTRAPOLATE_TIME = 'reflect'

LOOPS = 5  # loops within each run
BETWEEN_LOOPS = 5  # loops used to realign different runs
SPEEDUP = 5  # image sub-sampling factor for speeding up
"""
# How to tune those parameters for a multi-resolution implementation
LOOPS = 5, 1
BETWEEN_LOOPS = 5, 1
SPEEDUP = 5, 2
"""


def interp_slice_order(Z, slice_order):
    Z = np.asarray(Z)
    nslices = len(slice_order)
    aux = np.asarray(list(slice_order) + [slice_order[0] + nslices])
    Zf = np.floor(Z).astype('int')
    w = Z - Zf
    Zal = Zf % nslices
    Za = Zal + w
    ret = (1 - w) * aux[Zal] + w * aux[Zal + 1]
    ret += (Z - Za)
    return ret


def scanner_coords(xyz, affine, from_world, to_world):
    Tv = np.dot(from_world, np.dot(affine, to_world))
    XYZ = apply_affine(Tv, xyz)
    return XYZ[:, 0], XYZ[:, 1], XYZ[:, 2]


def make_grid(dims, subsampling=(1, 1, 1), borders=(0, 0, 0)):
    slices = [slice(b, d - b, s)\
                  for d, s, b in zip(dims, subsampling, borders)]
    xyz = np.mgrid[slices]
    xyz = np.rollaxis(xyz, 0, 4)
    xyz = np.reshape(xyz, [np.prod(xyz.shape[0:-1]), 3])
    return xyz


class Image4d(object):
    """
    Class to represent a sequence of 3d scans (possibly acquired on a
    slice-by-slice basis).

    Object remains empty until the data array is actually loaded in memory.

    Parameters
    ----------
      data : nd array or proxy (function that actually gets the array)
    """
    def __init__(self, data, affine, tr, tr_slices=None, start=0.0,
                 slice_order=SLICE_ORDER, interleaved=INTERLEAVED,
                 slice_info=None):
        """
        Configure fMRI acquisition time parameters.
        """
        self.affine = np.asarray(affine)
        self.tr = float(tr)
        self.start = float(start)
        self.interleaved = bool(interleaved)

        # guess the slice axis and direction (z-axis)
        if slice_info == None:
            orient = io_orientation(self.affine)
            self.slice_axis = int(np.where(orient[:, 0] == 2)[0])
            self.slice_direction = int(orient[self.slice_axis, 1])
        else:
            self.slice_axis = int(slice_info[0])
            self.slice_direction = int(slice_info[1])

        # unformatted parameters
        self._tr_slices = tr_slices
        self._slice_order = slice_order

        if isinstance(data, np.ndarray):
            self._data = data
            self._shape = data.shape
            self._get_data = None
            self._init_timing_parameters()
        else:
            self._data = None
            self._shape = None
            self._get_data = data

    def _load_data(self):
        self._data = self._get_data()
        self._shape = self._data.shape
        self._init_timing_parameters()

    def get_data(self):
        if self._data == None:
            self._load_data()
        return self._data
    
    def get_shape(self):
        if self._shape == None:
            self._load_data()
        return self._shape

    def _init_timing_parameters(self):
        # Number of slices
        nslices = self.get_shape()[self.slice_axis]
        self.nslices = nslices
        # Default slice repetition time (no silence)
        if self._tr_slices == None:
            self.tr_slices = self.tr / float(nslices)
        else:
            self.tr_slices = float(self._tr_slices)
        # Set slice order
        if isinstance(self._slice_order, str):
            if not self.interleaved:
                aux = range(nslices)
            else:
                aux = range(nslices)[0::2] + range(nslices)[1::2]
            if self._slice_order == 'descending':
                aux.reverse()
            self.slice_order = np.array(aux)
        else:
            # Verify correctness of provided slice indexes
            provided_slices = np.array(sorted(self._slice_order))
            if np.any(provided_slices != np.arange(nslices)):
                raise ValueError(
                    "Incorrect slice indexes were provided. There are %d "
                    "slices in the volume, indexes should start from 0 and "
                    "list all slices. "
                    "Provided slice_order: %s" % (nslices, self._slice_order))
            self.slice_order = np.asarray(self._slice_order)

    def z_to_slice(self, z):
        """
        Account for the fact that slices may be stored in reverse
        order wrt the scanner coordinate system convention (slice 0 ==
        bottom of the head)
        """
        if self.slice_direction < 0:
            return self.nslices - 1 - z
        else:
            return z

    def scanner_time(self, zv, t):
        """
        tv = scanner_time(zv, t)
        zv, tv are grid coordinates; t is an actual time value.
        """
        corr = self.tr_slices * interp_slice_order(self.z_to_slice(zv),
                                                   self.slice_order)
        return (t - self.start - corr) / self.tr

    def free_data(self):
        if not self._get_data == None:
            self._data = None


class Realign4dAlgorithm(object):

    def __init__(self,
                 im4d,
                 affine_class=Rigid,
                 transforms=None,
                 time_interp=True,
                 subsampling=(1, 1, 1),
                 borders=BORDERS,
                 optimizer=OPTIMIZER,
                 optimize_template=True,
                 xtol=XTOL,
                 ftol=FTOL,
                 gtol=GTOL,
                 stepsize=STEPSIZE,
                 maxiter=MAXITER,
                 maxfun=MAXFUN,
                 refscan=REFSCAN):

        self.dims = im4d.get_shape()
        self.nscans = self.dims[3]
        self.xyz = make_grid(self.dims[0:3], subsampling, borders)
        masksize = self.xyz.shape[0]
        self.data = np.zeros([masksize, self.nscans], dtype='double')

        # Initialize space/time transformation parameters
        self.affine = im4d.affine
        self.inv_affine = np.linalg.inv(self.affine)
        if transforms == None:
            self.transforms = [affine_class() for scan in range(self.nscans)]
        else:
            self.transforms = transforms
        self.scanner_time = im4d.scanner_time
        self.timestamps = im4d.tr * np.arange(self.nscans)

        # Compute the 4d cubic spline transform
        self.time_interp = time_interp
        if time_interp:
            self.cbspline = _cspline_transform(im4d.get_data())
        else:
            self.cbspline = np.zeros(self.dims, dtype='double')
            for t in range(self.dims[3]):
                self.cbspline[:, :, :, t] =\
                    _cspline_transform(im4d.get_data()[:, :, :, t])

        # The reference scan conventionally defines the head
        # coordinate system
        self.optimize_template = optimize_template
        if not optimize_template and refscan == None:
            self.refscan = REFSCAN
        else:
            self.refscan = refscan

        # Set the minimization method
        self.set_fmin(optimizer, stepsize,
                      xtol=xtol,
                      ftol=ftol,
                      gtol=gtol,
                      maxiter=maxiter,
                      maxfun=maxfun)

        # Auxiliary array for realignment estimation
        self._res = np.zeros(masksize, dtype='double')
        self._res0 = np.zeros(masksize, dtype='double')
        self._aux = np.zeros(masksize, dtype='double')
        self.A = np.zeros((masksize, self.transforms[0].param.size),
                          dtype='double')
        self._pc = None

    def resample(self, t):
        """
        Resample a particular time frame on the (sub-sampled) working
        grid.

        x,y,z,t are "head" grid coordinates
        X,Y,Z,T are "scanner" grid coordinates
        """
        X, Y, Z = scanner_coords(self.xyz, self.transforms[t].as_affine(),
                                 self.inv_affine, self.affine)
        if self.time_interp:
            T = self.scanner_time(Z, self.timestamps[t])
            _cspline_sample4d(self.data[:, t],
                              self.cbspline,
                              X, Y, Z, T,
                              mx=EXTRAPOLATE_SPACE,
                              my=EXTRAPOLATE_SPACE,
                              mz=EXTRAPOLATE_SPACE,
                              mt=EXTRAPOLATE_TIME)
        else:
            _cspline_sample3d(self.data[:, t],
                              self.cbspline[:, :, :, t],
                              X, Y, Z,
                              mx=EXTRAPOLATE_SPACE,
                              my=EXTRAPOLATE_SPACE,
                              mz=EXTRAPOLATE_SPACE)

    def resample_full_data(self):
        if VERBOSE:
            print('Gridding...')
        xyz = make_grid(self.dims[0:3])
        res = np.zeros(self.dims)
        for t in range(self.nscans):
            if VERBOSE:
                print('Fully resampling scan %d/%d' % (t + 1, self.nscans))
            X, Y, Z = scanner_coords(xyz, self.transforms[t].as_affine(),
                                     self.inv_affine, self.affine)
            if self.time_interp:
                T = self.scanner_time(Z, self.timestamps[t])
                _cspline_sample4d(res[:, :, :, t],
                                  self.cbspline,
                                  X, Y, Z, T,
                                  mt='nearest')
            else:
                _cspline_sample3d(res[:, :, :, t],
                                  self.cbspline[:, :, :, t],
                                  X, Y, Z)
        return res

    def set_fmin(self, optimizer, stepsize, **kwargs):
        """
        Return the minimization function
        """
        self.stepsize = stepsize
        self.optimizer = optimizer
        self.optimizer_kwargs = kwargs
        self.optimizer_kwargs.setdefault('xtol', XTOL)
        self.optimizer_kwargs.setdefault('ftol', FTOL)
        self.optimizer_kwargs.setdefault('gtol', GTOL)
        self.optimizer_kwargs.setdefault('maxiter', MAXITER)
        self.optimizer_kwargs.setdefault('maxfun', MAXFUN)
        self.use_derivatives = use_derivatives(self.optimizer)

    def init_instant_motion(self, t):
        """
        Pre-compute and cache some constants (at fixed time) for
        repeated computations of the alignment energy.

        The idea is to decompose the average temporal variance via:

        V = (n-1)/n V* + (n-1)/n^2 (x-m*)^2

        with x the considered volume at time t, and m* the mean of all
        resampled volumes but x. Only the second term is variable when

        one volumes while the others are fixed. A similar
        decomposition is used for the global variance, so we end up
        with:

        V/V0 = [nV* + (x-m*)^2] / [nV0* + (x-m0*)^2]
        """
        fixed = range(self.nscans)
        fixed.remove(t)
        aux = self.data[:, fixed]
        if self.optimize_template:
            self.mu = np.mean(aux, 1)
        self.offset = self.nscans * np.mean((aux.T - self.mu) ** 2)
        self.mu0 = np.mean(aux)
        self.offset0 = self.nscans * np.mean((aux - self.mu0) ** 2)
        self._t = t
        self._pc = None

    def set_transform(self, t, pc):
        self.transforms[t].param = pc
        self.resample(t)

    def _init_energy(self, pc):
        if pc is self._pc:
            return
        self.set_transform(self._t, pc)
        self._pc = pc
        self._res[:] = self.data[:, self._t] - self.mu[:]
        self._V = np.maximum(self.offset + np.mean(self._res ** 2), SMALL)
        self._res0[:] = self.data[:, self._t] - self.mu0
        self._V0 = np.maximum(self.offset0 + np.mean(self._res0 ** 2), SMALL)

        if self.use_derivatives:
            # linearize the data wrt the transform parameters
            # use the auxiliary array to save the current resampled data
            self._aux[:] = self.data[:, self._t]
            basis = np.eye(6)
            for j in range(pc.size):
                self.set_transform(self._t, pc + self.stepsize * basis[j])
                self.A[:, j] = (self.data[:, self._t] - self._aux)\
                    / self.stepsize
            self.transforms[self._t].param = pc
            self.data[:, self._t] = self._aux[:]
            # pre-compute gradient and hessian of numerator and
            # denominator
            c = 2 / float(self.data.shape[0])
            self._dV = c * np.dot(self.A.T, self._res)
            self._dV0 = c * np.dot(self.A.T, self._res0)
            self._H = c * np.dot(self.A.T, self.A)

    def _energy(self):
        """
        The alignment energy is defined as the log-ratio between the
        average temporal variance in the sequence and the global
        spatio-temporal variance.
        """
        return np.log(self._V / self._V0)

    def _energy_gradient(self):
        return self._dV / self._V - self._dV0 / self._V0

    def _energy_hessian(self):
        return (1 / self._V - 1 / self._V0) * self._H\
            - np.dot(self._dV, self._dV.T) / np.maximum(self._V ** 2, SMALL)\
            + np.dot(self._dV0, self._dV0.T) / np.maximum(self._V0 ** 2, SMALL)

    def estimate_instant_motion(self, t):
        """
        Estimate motion parameters at a particular time.
        """
        if VERBOSE:
            print('Estimating motion at time frame %d/%d...'
                  % (t + 1, self.nscans))

        def f(pc):
            self._init_energy(pc)
            return self._energy()

        def fprime(pc):
            self._init_energy(pc)
            return self._energy_gradient()

        def fhess(pc):
            self._init_energy(pc)
            return self._energy_hessian()

        self.init_instant_motion(t)
        fmin, args, kwargs =\
            configure_optimizer(self.optimizer,
                                fprime=fprime,
                                fhess=fhess,
                                **self.optimizer_kwargs)

        # With scipy >= 0.9, some scipy minimization functions like
        # fmin_bfgs may crash due to the subroutine
        # `scalar_search_armijo` returning None as a stepsize when
        # unhappy about the objective function. This seems to have the
        # potential to occur in groupwise registration when using
        # strong image subsampling, i.e. at the coarser levels of the
        # multiscale pyramid. To avoid crashes, we insert a try/catch
        # instruction.
        try:
            pc = fmin(f, self.transforms[t].param, disp=VERBOSE, *args, **kwargs)
            self.set_transform(t, pc)
        except:
            warnings.warn('Minimization failed')

    def estimate_motion(self):
        """
        Optimize motion parameters for the whole sequence. All the
        time frames are initially resampled according to the current
        space/time transformation, the parameters of which are further
        optimized sequentially.
        """
        for t in range(self.nscans):
            if VERBOSE:
                print('Resampling scan %d/%d' % (t + 1, self.nscans))
            self.resample(t)

        # Set the template as the reference scan (will be overwritten
        # if template is to be optimized)
        if not hasattr(self, 'template'):
            self.mu = self.data[:, self.refscan].copy()
        for t in range(self.nscans):
            self.estimate_instant_motion(t)
            if VERBOSE:
                print(self.transforms[t])

    def align_to_refscan(self):
        """
        The `motion_estimate` method aligns scans with an online
        template so that spatial transforms map some average head
        space to the scanner space. To conventionally redefine the
        head space as being aligned with some reference scan, we need
        to right compose each head_average-to-scanner transform with
        the refscan's 'to head_average' transform.
        """
        if self.refscan == None:
            return
        Tref_inv = self.transforms[self.refscan].inv()
        for t in range(self.nscans):
            self.transforms[t] = (self.transforms[t]).compose(Tref_inv)


def resample4d(im4d, transforms, time_interp=True):
    """
    Resample a 4D image according to the specified sequence of spatial
    transforms, using either 4D interpolation if `time_interp` is True
    and 3D interpolation otherwise.
    """
    r = Realign4dAlgorithm(im4d, transforms=transforms,
                           time_interp=time_interp)
    res = r.resample_full_data()
    im4d.free_data()
    return res


def adjust_subsampling(speedup, dims):
    dims = np.array(dims)
    aux = np.maximum(speedup * dims / np.prod(dims) ** (1 / 3.), [1, 1, 1])
    return aux.astype('int')


def single_run_realign4d(im4d,
                         affine_class=Rigid,
                         time_interp=True,
                         loops=LOOPS,
                         speedup=SPEEDUP,
                         borders=BORDERS,
                         optimizer=OPTIMIZER,
                         xtol=XTOL,
                         ftol=FTOL,
                         gtol=GTOL,
                         stepsize=STEPSIZE,
                         maxiter=MAXITER,
                         maxfun=MAXFUN,
                         refscan=REFSCAN):
    """
    Realign a single run in space and time.

    Parameters
    ----------
    im4d : Image4d instance

    speedup : int or sequence
      If a sequence, implement a multi-scale

    """
    if not type(loops) in (list, tuple, np.array):
        loops = [loops]
    repeats = len(loops)

    def format_arg(x):
        if not type(x) in (list, tuple, np.array):
            x = [x for i in range(repeats)]
        else:
            if not len(x) == repeats:
                raise ValueError('inconsistent length in arguments')
        return x

    speedup = format_arg(speedup)
    optimizer = format_arg(optimizer)
    xtol = format_arg(xtol)
    ftol = format_arg(ftol)
    gtol = format_arg(gtol)
    stepsize = format_arg(stepsize)
    maxiter = format_arg(maxiter)
    maxfun = format_arg(maxfun)

    transforms = None
    opt_params = zip(loops, speedup, optimizer,
                     xtol, ftol, gtol,
                     stepsize, maxiter, maxfun)

    for loops_, speedup_, optimizer_, xtol_, ftol_, gtol_,\
            stepsize_, maxiter_, maxfun_ in opt_params:
        subsampling = adjust_subsampling(speedup_, im4d.get_shape()[0:3])

        r = Realign4dAlgorithm(im4d,
                               transforms=transforms,
                               affine_class=affine_class,
                               time_interp=time_interp,
                               subsampling=subsampling,
                               borders=borders,
                               refscan=refscan,
                               optimizer=optimizer_,
                               xtol=xtol_,
                               ftol=ftol_,
                               gtol=gtol_,
                               stepsize=stepsize_,
                               maxiter=maxiter_,
                               maxfun=maxfun_)

        for loop in range(loops_):
            r.estimate_motion()

        r.align_to_refscan()
        transforms = r.transforms
        im4d.free_data()

    return transforms


def realign4d(runs,
              affine_class=Rigid,
              time_interp=True,
              align_runs=True,
              loops=LOOPS,
              between_loops=BETWEEN_LOOPS,
              speedup=SPEEDUP,
              borders=BORDERS,
              optimizer=OPTIMIZER,
              xtol=XTOL,
              ftol=FTOL,
              gtol=GTOL,
              stepsize=STEPSIZE,
              maxiter=MAXITER,
              maxfun=MAXFUN,
              refscan=REFSCAN):
    """
    Parameters
    ----------

    runs : list of Image4d objects

    Returns
    -------
    transforms : list
                 nested list of rigid transformations


    transforms map an 'ideal' 4d grid (conventionally aligned with the
    first scan of the first run) to the 'acquisition' 4d grid for each
    run
    """

    # Single-session case
    if not type(runs) in (list, tuple, np.array):
        runs = [runs]
    nruns = len(runs)
    if nruns == 1:
        align_runs = False

    # Correct motion and slice timing in each sequence separately
    transforms = [single_run_realign4d(run,
                                       affine_class=affine_class,
                                       time_interp=time_interp,
                                       loops=loops,
                                       speedup=speedup,
                                       borders=borders,
                                       optimizer=optimizer,
                                       xtol=xtol,
                                       ftol=ftol,
                                       gtol=gtol,
                                       stepsize=stepsize,
                                       maxiter=maxiter,
                                       maxfun=maxfun,
                                       refscan=refscan) for run in runs]
    if not align_runs:
        return transforms, transforms, None

    # Correct between-session motion using the mean image of each
    # corrected run, and creating a fake time series with no temporal
    # smoothness
    ## FIXME: check that all runs have the same to-world transform
    mean_img_shape = list(runs[0].get_shape()[0:3]) + [nruns]
    mean_img_data = np.zeros(mean_img_shape)

    for i in range(nruns):
        corr_run = resample4d(runs[i], transforms=transforms[i],
                              time_interp=time_interp)
        mean_img_data[..., i] = corr_run.mean(3)
    del corr_run

    mean_img = Image4d(mean_img_data, affine=runs[0].affine,
                       tr=1.0, tr_slices=0.0)
    transfo_mean = single_run_realign4d(mean_img,
                                        affine_class=affine_class,
                                        time_interp=False,
                                        loops=between_loops,
                                        speedup=speedup,
                                        borders=borders,
                                        optimizer=optimizer,
                                        xtol=xtol,
                                        ftol=ftol,
                                        gtol=gtol,
                                        stepsize=stepsize,
                                        maxiter=maxiter,
                                        maxfun=maxfun)

    # Compose transformations for each run
    ctransforms = [None for i in range(nruns)]
    for i in range(nruns):
        ctransforms[i] = [t.compose(transfo_mean[i]) for t in transforms[i]]
    return ctransforms, transforms, transfo_mean


class Realign4d(object):

    def __init__(self, images, affine_class=Rigid):
        self._generic_init(images, affine_class, SLICE_ORDER, INTERLEAVED,
                           1.0, 0.0, 0.0, False, None)

    def _generic_init(self, images, affine_class,
                      slice_order, interleaved, tr, tr_slices,
                      start, time_interp, slice_info):
        if slice_order == None:
            slice_order = SLICE_ORDER
            if time_interp:
                raise ValueError('Slice order is requested'
                          + ' with time interpolation switched on')
            time_interp = False
        if tr == None:
            raise ValueError('Repetition time cannot be None')
        if not type(images) in (list, tuple, np.array):
            images = [images]
        self._runs = []
        self.affine_class = affine_class
        for im in images:
            xyz_img = as_xyz_image(im)
            self._runs.append(Image4d(xyz_img.get_data,
                                      xyz_affine(xyz_img),
                                      tr=tr, tr_slices=tr_slices,
                                      start=start, slice_order=slice_order,
                                      interleaved=interleaved,
                                      slice_info=slice_info))
        self._transforms = [None for run in self._runs]
        self._within_run_transforms = [None for run in self._runs]
        self._mean_transforms = [None for run in self._runs]
        self._time_interp = time_interp

    def estimate(self,
                 loops=LOOPS,
                 between_loops=None,
                 align_runs=True,
                 speedup=SPEEDUP,
                 borders=BORDERS,
                 optimizer=OPTIMIZER,
                 xtol=XTOL,
                 ftol=FTOL,
                 gtol=GTOL,
                 stepsize=STEPSIZE,
                 maxiter=MAXITER,
                 maxfun=MAXFUN,
                 refscan=REFSCAN):
        if between_loops == None:
            between_loops = loops
        t = realign4d(self._runs,
                      affine_class=self.affine_class,
                      time_interp=self._time_interp,
                      align_runs=align_runs,
                      loops=loops,
                      between_loops=between_loops,
                      speedup=speedup,
                      borders=borders,
                      optimizer=optimizer,
                      xtol=xtol,
                      ftol=ftol,
                      gtol=gtol,
                      stepsize=stepsize,
                      maxiter=maxiter,
                      maxfun=maxfun,
                      refscan=refscan)
        self._transforms, self._within_run_transforms,\
            self._mean_transforms = t

    def resample(self, r=None, align_runs=True):
        """
        Return the resampled run number r as a 4d nipy-like
        image. Returns all runs as a list of images if r == None.
        """
        if align_runs:
            transforms = self._transforms
        else:
            transforms = self._within_run_transforms
        runs = range(len(self._runs))
        if r == None:
            data = [resample4d(self._runs[r], transforms=transforms[r],
                               time_interp=self._time_interp) for r in runs]
            return [make_xyz_image(data[r], self._runs[r].affine, 'scanner')
                    for r in runs]
        else:
            data = resample4d(self._runs[r], transforms=transforms[r],
                              time_interp=self._time_interp)
            return make_xyz_image(data, self._runs[r].affine, 'scanner')


class FmriRealign4d(Realign4d):

    def __init__(self, images, slice_order, interleaved=None,
                 tr=1.0, tr_slices=None, start=0.0, time_interp=True,
                 affine_class=Rigid, slice_info=None):

        """
        Spatiotemporal realignment class for fMRI series.

        Parameters
        ----------
        images : image or list of images
          Single or multiple input 4d images representing one or
          several fMRI runs.

        tr : float
          Inter-scan repetition time, i.e. the time elapsed between
          two consecutive scans. The unit in which `tr` is given is
          arbitrary although it needs to be consistent with the
          `tr_slices` and `start` arguments.

        tr_slices : float
          Inter-slice repetition time, same as tr for slices. If None,
          acquisition is assumed continuous and `tr_slices` is set to
          `tr` divided by the number of slices.

        start : float
          Starting acquisition time respective to the implicit time
          origin.

        slice_order : str or array-like
          If str, one of {'ascending', 'descending'}. If array-like,
          then the order in which the slices were collected in
          time. For instance, the following represents an ascending
          contiguous sequence:

          slice_order = [0, 1, 2, ...]

        interleaved : bool
          Deprecated.

          Whether slice acquisition order is interleaved. Ignored if
          `slice_order` is array-like.

          If slice_order=='ascending' and interleaved==True, the
          assumed slice order is:

          [0, 2, 4, ..., 1, 3, 5, ...]

          If slice_order=='descending' and interleaved==True, the
          assumed slice order is:

          [N-1, N-3, N-5, ..., N-2, N-4, N-6]

          Given that there exist other types of interleaved
          acquisitions depending on scanner settings and
          manufacturers, it is strongly recommended to input the
          slice_order as an array unless you are sure what you are
          doing.

        slice_info : None or tuple, optional
          None, or a tuple with slice axis as the first element and
          direction as the second, for instance (2, 1).  If None, then
          guess the slice axis, and direction, as the closest to the z
          axis, as estimated from the affine.
        """
        if not interleaved == None:
            warnings.warn('interleaved keyword is deprecated. Please input explicit slice order instead.')
        self._generic_init(images, affine_class, slice_order, interleaved,
                           tr, tr_slices, start, time_interp, slice_info)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import with_statement

import gc
import warnings

import numpy as np

from nipy.modalities.fmri.api import axis0_generator, FmriImageList
from nipy.core.api import parcels, Image, AffineTransform as AfT
from nipy.io.api import  load_image, save_image

from nose.tools import assert_equal, assert_true

from nibabel.tmpdirs import InTemporaryDirectory
from nipy.testing import funcfile


def setup():
    # Suppress warnings during tests to reduce noise
    warnings.simplefilter("ignore")


def teardown():
    # Clear list of warning filters
    warnings.resetwarnings()


def test_write():
    fname = 'myfile.nii'
    img = load_image(funcfile)
    with InTemporaryDirectory():
        save_image(img, fname)
        test = FmriImageList.from_image(load_image(fname))
        assert_equal(test[0].affine.shape, (4,4))
        assert_equal(img[0].affine.shape, (5,4))
        # Check the affine...
        A = np.identity(4)
        A[:3,:3] = img[:,:,:,0].affine[:3,:3]
        A[:3,-1] = img[:,:,:,0].affine[:3,-1]
        assert_true(np.allclose(test[0].affine, A))
        del test


def test_iter():
    img = load_image(funcfile)
    img_shape = img.shape
    exp_shape = (img_shape[0],) + img_shape[2:]
    j = 0
    for i, d in axis0_generator(img.get_data()):
        j += 1
        assert_equal(d.shape, exp_shape)
        del(i); gc.collect()
    assert_equal(j, img_shape[1])


def test_subcoordmap():
    img = load_image(funcfile)
    subcoordmap = img[3].coordmap
    xform = img.affine[:,1:]
    assert_true(np.allclose(subcoordmap.affine[1:], xform[1:]))
    assert_true(np.allclose(subcoordmap.affine[0], [0,0,0,img.coordmap([3,0,0,0])[0]]))


def test_labels1():
    img = load_image(funcfile)
    data = img.get_data()
    parcelmap = Image(img[0].get_data(), AfT('kji', 'zyx', np.eye(4)))
    parcelmap = (parcelmap.get_data() * 100).astype(np.int32)
    v = 0
    for i, d in axis0_generator(data, parcels(parcelmap)):
        v += d.shape[1]
    assert_equal(v, parcelmap.size)



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function, with_statement

DESCRIP = 'Run and log examples'
EPILOG = \
""" Run examples in directory

Typical usage is:

run_log_examples.py nipy/examples --log-path=~/tmp/eg_logs

to run the examples and log the result, or

run_log_examples.py nipy/examples/some_example.py

to run a single example.
"""

import sys
import os
from os.path import (abspath, expanduser, join as pjoin, sep as psep, isfile,
                     dirname)
from subprocess import Popen, PIPE
import re

from nibabel.py3k import asstr

from nipy.externals.argparse import (ArgumentParser,
                                     RawDescriptionHelpFormatter)


PYTHON=sys.executable
NEED_SHELL = True

class ProcLogger(object):
    def __init__(self, log_path, working_path):
        self.log_path = log_path
        self.working_path = working_path
        self._names = []

    def cmd_str_maker(self, cmd, args):
        return " ".join([cmd] + list(args))

    def __call__(self, cmd_name, cmd, args=(), cwd=None):
        # Mqke log files
        if cmd_name in self._names:
            raise ValueError('Command name {0} not unique'.format(cmd_name))
        self._names.append(cmd_name)
        if cwd is None:
            cwd = self.working_path
        cmd_out_path = pjoin(self.log_path, cmd_name)
        stdout_log = open(cmd_out_path + '.stdout', 'wt')
        stderr_log = open(cmd_out_path + '.stderr', 'wt')
        try:
            # Start subprocess
            cmd_str = self.cmd_str_maker(cmd, args)
            proc = Popen(cmd_str,
                        cwd = cwd,
                        stdout = stdout_log,
                        stderr = stderr_log,
                        shell = NEED_SHELL)
            # Execute
            retcode = proc.wait()
        finally:
            if proc.poll() is None: # In case we get killed
                proc.terminate()
            stdout_log.close()
            stderr_log.close()
        return retcode

    def run_pipes(self, cmd, args=(), cwd=None):
        if cwd is None:
            cwd = self.working_path
        try:
            # Start subprocess
            cmd_str = self.cmd_str_maker(cmd, args)
            proc = Popen(cmd_str,
                         cwd = cwd,
                         stdout = PIPE,
                         stderr = PIPE,
                         shell = NEED_SHELL)
            # Execute
            stdout, stderr = proc.communicate()
        finally:
            if proc.poll() is None: # In case we get killed
                proc.terminate()
        return asstr(stdout), asstr(stderr), proc.returncode


class PyProcLogger(ProcLogger):
    def cmd_str_maker(self, cmd, args):
        """ Execute python script `cmd`

        Reject any `args` because we're using ``exec`` to execute the script.

        Prepend some matplotlib setup to suppress figures
        """
        if len(args) != 0:
            raise ValueError("Cannot use args with {8}".format(self.__class__))
        return("""{0} -c "import matplotlib as mpl; mpl.use('agg'); """
               """exec(open('{1}', 'rt').read())" """.format(PYTHON, cmd))


def _record(result, fname, fileobj):
    print(result)
    fileobj.write('{0}: {1}\n'.format(fname, result))


def main():
    parser = ArgumentParser(description=DESCRIP,
                            epilog=EPILOG,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('examples_path', type=str,
                        help='filename of example or directory containing '
                             'examples to run')
    parser.add_argument('--log-path', type=str, default='',
                        help='path for output logs (default is cwd)')
    parser.add_argument('--excludex', type=str, action='append', default=[],
                        help='regex for files to exclude (add more than one '
                        '--excludex option for more than one regex filter')
    args = parser.parse_args()
    # Proc runner
    eg_path = abspath(expanduser(args.examples_path))
    if args.log_path == '':
        log_path = abspath(os.getcwd())
    else:
        log_path = abspath(expanduser(args.log_path))
    excludexes = [re.compile(s) for s in args.excludex]
    if isfile(eg_path): # example was a file
        proc_logger = PyProcLogger(log_path=log_path,
                                   working_path=dirname(eg_path))
        print("Running " + eg_path)
        stdout, stderr, code = proc_logger.run_pipes(eg_path)
        print('==== Stdout ====')
        print(stdout)
        print('==== Stderr ====')
        print(stderr)
        sys.exit(code)
    # Multi-run with logging to file
    proc_logger = PyProcLogger(log_path=log_path,
                               working_path=eg_path)
    fails = 0
    with open(pjoin(log_path, 'summary.txt'), 'wt') as f:
        for dirpath, dirnames, filenames in os.walk(eg_path):
            for fname in filenames:
                full_fname = pjoin(dirpath, fname)
                if fname.endswith(".py"):
                    print(fname, end=': ')
                    sys.stdout.flush()
                    for excludex in excludexes:
                        if excludex.search(fname):
                            _record('SKIP', fname, f)
                            break
                    else: # run test
                        cmd_name = full_fname.replace(eg_path + psep, '')
                        cmd_name = cmd_name.replace(psep, '-')
                        code = proc_logger(cmd_name, full_fname, cwd=dirpath)
                        if code == 0:
                            _record('OK', fname, f)
                        else:
                            fails += 1
                            _record('FAIL', fname, f)
    sys.exit(fails if fails < 255 else 255)


if __name__ == '__main__':
    main()
